/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import junit.framework.TestCase;

import org.ow2.bonita.facade.CommandAPI;
import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.QueryDefinitionAPI;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.impl.StandardInternalAPIAccessorFactory;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.runtime.var.Enumeration;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.BusinessArchiveFactory;
import org.ow2.novabpm.util.ClassDataTool;
import org.ow2.novabpm.util.Misc;
import org.ow2.novabpm.util.SimpleCallbackHandler;

public abstract class APITestCase extends TestCase {

  static {
    // Initialize the Jaas login configuration with a default value
    // This works only for running tests from eclipse
    final String defaultLoginFile = "modules/distrib/resources/conf/jaas-standard.cfg";
    final String jaasKey = "java.security.auth.login.config";
    final String loginFile = System.getProperty(jaasKey, defaultLoginFile);
    if (loginFile.equals(defaultLoginFile)) {
      System.setProperty(jaasKey, defaultLoginFile);
    }
    final String defaultLoggingFile = "modules/distrib/resources/conf/logging.properties";
    final String loggingKey = "java.util.logging.config.file";
    final String loggingFile = System.getProperty(loggingKey, defaultLoggingFile);
    if (loggingFile.equals(defaultLoggingFile)) {
      System.setProperty(loggingKey, defaultLoggingFile);
    }
  }

  /** LOG. */
  protected static final Logger LOG = Logger.getLogger(APITestCase.class.getName());
  private RuntimeAPI runtimeAPI;
  private ManagementAPI managementAPI;
  private QueryRuntimeAPI queryRuntimeAPI;
  private QueryDefinitionAPI queryDefinitionAPI;
  private CommandAPI commandAPI;

  protected LoginContext loginContext;

  protected static final String STANDALONE_ACCESS = "standalone_access";
  protected static final String EJB_ACCESS = "ejb_access";

  // This boolean is set to false is the previous test was successful,
  // this way, the db is recreated only when the previous test has failed
  private static boolean recreateDb = false;

  private boolean testFail = false;

  @Override
  protected void setUp() throws Exception {
    super.setUp();
    if (LOG.isLoggable(Level.WARNING)) {
      LOG.warning("======== Starting test: " + this.getClass().getName() + "." + this.getName() + "() ==========");
    }
    login();

    if (isLocalTest()) {
      LocalTestSetup.setUp(recreateDb);
    }
  }

  @Override
  protected void tearDown() throws Exception {
    if (isLocalTest()) {
      LocalTestSetup.tearDown(this.testFail);
    }
    if (this.loginContext != null) {
      this.loginContext.logout();
    }
    if (LOG.isLoggable(Level.WARNING)) {
      LOG.warning("======== Ending test: " + this.getName() + "==========");
    }
    //LOG = null;
    this.runtimeAPI = null;
    this.managementAPI = null;
    this.queryRuntimeAPI = null;
    this.queryDefinitionAPI = null;
    this.commandAPI = null;
    this.loginContext = null;
  }


  private boolean isLocalTest() {
    try {
      StandardInternalAPIAccessorFactory.getStandardInternalAPIAccessor();
      return true;
    } catch (final BonitaRuntimeException e) {
      return false;
    }
  }

  @Override
  protected void runTest() throws Throwable {
    try {
      super.runTest();
      // All went well, don't need to recreate db
      recreateDb = false;
    } catch (final Throwable e) {
      // There is a failure, we need to recreate db
      if (isLocalTest()) {
        recreateDb = true;
      }
      this.testFail = true;
      throw e;
    }
  }

  /*
   * APIs
   */
  protected synchronized RuntimeAPI getRuntimeAPI() {
    if (this.runtimeAPI == null) {
      this.runtimeAPI = AccessorUtil.getRuntimeAPI();
    }
    return this.runtimeAPI;
  }

  protected synchronized ManagementAPI getManagementAPI() {
    if (this.managementAPI == null) {
      this.managementAPI = AccessorUtil.getManagementAPI();
    }
    return this.managementAPI;
  }

  protected synchronized QueryRuntimeAPI getQueryRuntimeAPI() {
    if (this.queryRuntimeAPI == null) {
      this.queryRuntimeAPI = AccessorUtil.getQueryRuntimeAPI();
    }
    return this.queryRuntimeAPI;
  }

  protected synchronized QueryDefinitionAPI getQueryDefinitionAPI() {
    if (this.queryDefinitionAPI == null) {
      this.queryDefinitionAPI = AccessorUtil.getQueryDefinitionAPI();
    }
    return this.queryDefinitionAPI;
  }

  protected synchronized CommandAPI getCommandAPI() {
    if (this.commandAPI == null) {
      this.commandAPI = AccessorUtil.getCommandAPI();
    }
    return this.commandAPI;
  }


  /*
   * BEGIN OF TO BE REMOVED
   */
  protected void checkExecutedOnce(final ProcessInstance instance,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    checkExecutedOnce(instance.getUUID(), activityIds);
  }
  protected void checkExecutedOnce(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    checkExecuted(instanceUUID, activityIds);
    final ProcessInstance processInstance = getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    assertEquals(InstanceState.COMPLETED, processInstance.getInstanceState());
  }

  protected void checkNotExecuted(final ProcessInstance processInstance,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    assertEquals(InstanceState.COMPLETED, processInstance.getInstanceState());
    for (final String activityId : activityIds) {
      try {
        getQueryRuntimeAPI().getActivityInstances(processInstance.getUUID(), activityId);
        fail("Activity : " + activityId + " must not be executed in this instance!");
      } catch (final ActivityNotFoundException e) {
        //nothing, we expect an exception
      }
    }
  }

  protected void checkStopped(final ProcessInstance instance,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    checkStopped(instance.getUUID(), activityIds);
  }
  protected void checkStopped(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    final ProcessInstance instance = getQueryRuntimeAPI().getProcessInstance(instanceUUID);
    assertNotNull("Can't find an instance with uuid : " + instanceUUID, instance);
    checkExecuted(instanceUUID, activityIds);
  }

  private void checkExecuted(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    for (final String activityId : activityIds) {
      final Set<ActivityInstance<ActivityBody>> activityInsts =
        getQueryRuntimeAPI().getActivityInstances(instanceUUID, activityId);
      assertEquals("Activity " + activityId + " executed more than once.", 1, activityInsts.size());
      final ActivityInstance<ActivityBody> activityInst = activityInsts.iterator().next();
      assertNotNull("Activity " + activityId + " not executed", activityInst);
      assertNotNull("Bad state for activity " + activityId, activityInst.getBody().getEndedDate());
    }
  }

  protected void checkExecutedManyTimes(final ProcessInstance instance,
      final String[] activityIds, final int execNb) throws InstanceNotFoundException, ActivityNotFoundException {
    checkExecutedManyTimes(instance.getUUID(), activityIds, execNb);
  }
  protected void checkExecutedManyTimes(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds, final int execNb) throws InstanceNotFoundException, ActivityNotFoundException {
    for (final String activityId : activityIds) {
      final Set<ActivityInstance<ActivityBody>> activities = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activityId);
      assertNotNull("Activity " + activityId + " not executed", activities);
      assertTrue("Activity " + activityId + " not executed", activities.size() > 0);
      assertEquals("Wrong number of executions for activity : " + activityId, execNb, activities.size());
      for (final ActivityInstance<ActivityBody> activityInst : activities) {
        assertNotNull("Bad state for activity " + activityId, activityInst.getBody().getStartedDate());
        assertNotNull("Bad state for activity " + activityId, activityInst.getBody().getEndedDate());
      }
    }
  }

  protected void checkOnlyOneExecuted(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException, ActivityNotFoundException {
    checkAtLeastOneExecuted(instanceUUID, activityIds);
    checkAtMostOneExecuted(instanceUUID, activityIds);
  }

  protected void checkAtMostOneExecuted(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException {
    boolean found = false;
    for (final String activityId : activityIds) {
      Set<ActivityInstance<ActivityBody>> activityInsts = null;
      try {
        activityInsts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activityId);
        for (final ActivityInstance<ActivityBody> activityInst : activityInsts) {
          if (activityInst.getBody().getEndedDate() != null) {
            assertFalse("More than one activity in "
                + Misc.componentsToString(activityIds, false)
                + " has been executed.", found);
            found = true;
          }
        }
      } catch (final ActivityNotFoundException e) {
        // nothing to do
      }

    }
  }

  protected void checkAtLeastOneExecuted(final ProcessInstanceUUID instanceUUID,
      final String[] activityIds) throws InstanceNotFoundException {
    for (final String activityId : activityIds) {
      try {
        final Set<ActivityInstance<ActivityBody>> activityInsts =
          getQueryRuntimeAPI().getActivityInstances(instanceUUID, activityId);
        for (final ActivityInstance<ActivityBody> activityInst : activityInsts) {
          if (activityInst.getBody().getEndedDate() != null) {
            return;
          }
        }
      } catch (final ActivityNotFoundException e) {
        // nothing to do
      }
    }
    fail("At least one activity of " + Misc.componentsToString(activityIds, false)
        + " has to be executed, none was found.");
  }

  /*
   * END OF TO BE REMOVED
   */

  @SuppressWarnings("null")
  protected void executeTask(final ProcessInstanceUUID instanceUUID, final String activityId) throws InstanceNotFoundException,
  TaskNotFoundException, IllegalTaskStateException {
    final Collection<ActivityInstance< TaskInstance >> taskActivities =
      getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);

    final Collection<TaskUUID> tasksToDo = new ArrayList<TaskUUID>();

    ActivityInstance< TaskInstance > matchingActivity = null;

    for (final ActivityInstance< TaskInstance > taskActivity : taskActivities) {
      final TaskInstance task = taskActivity.getBody();
      tasksToDo.add(task.getUUID());
      if (taskActivity.getProcessInstanceUUID().equals(instanceUUID)
          && taskActivity.getActivityId().equals(activityId)) {
        matchingActivity = taskActivity;
        break;
      }
    }
    assertFalse("No tasks in instance " + instanceUUID + " to do for the current user.", tasksToDo.isEmpty());

    assertNotNull("no activity with activityId : " + activityId + " is defined within instance : " + instanceUUID
        + " for the current user. Available tasks: " + taskActivities, matchingActivity);

    final TaskUUID taskUUID = matchingActivity.getBody().getUUID();

    assertNull(matchingActivity.getBody().getStartedDate());
    assertNull(matchingActivity.getBody().getEndedDate());

    assertEquals(TaskState.READY, matchingActivity.getBody().getState());

    getRuntimeAPI().startTask(taskUUID, true);

    matchingActivity = getQueryRuntimeAPI().getTask(taskUUID);
    assertEquals(TaskState.EXECUTING, matchingActivity.getBody().getState());

    assertNotNull(matchingActivity.getBody().getStartedDate());
    assertNull(matchingActivity.getBody().getEndedDate());

    getRuntimeAPI().finishTask(taskUUID, true);

    matchingActivity = getQueryRuntimeAPI().getTask(taskUUID);
    assertEquals(TaskState.FINISHED, matchingActivity.getBody().getState());

    assertNotNull(matchingActivity.getBody().getStartedDate());
    assertNotNull(matchingActivity.getBody().getEndedDate());

    final long startTime = matchingActivity.getBody().getStartedDate().getTime();
    final long endTime = matchingActivity.getBody().getEndedDate().getTime();
    assertTrue(endTime >= startTime);
  }



  /**
   *   check setup of enumeration variable
   */
  public static void checkEnumerationVariable(final Enumeration enu, final String selected, final String[] possibles) {
    assertNotNull(enu);
    if (enu != null) {
      final Set<String> enumValues = enu.getPossibleValues();
      assertEquals(Misc.deepToString(enu), selected, enu.getSelectedValue());
      assertTrue(enumValues.toString(), enumValues.containsAll(Arrays.asList(possibles)));
    }
  }

  protected boolean causeContains(final String message, final Throwable e) {
    if (e == null) {
      return false;
    }
    if (e.getMessage() != null && e.getMessage().contains(message)) {
      return true;
    }
    if (e.getCause() != null) {
      return causeContains(message, e.getCause());
    }
    return false;
  }

  /*
   * LOGIN
   */

  protected String getLogin() {
    return "admin";
  }
  protected String getPassword() {
    return "bpm";
  }
  protected void login() throws LoginException {
    this.loginContext = new LoginContext("Bonita", new SimpleCallbackHandler(getLogin(), getPassword()));
    this.loginContext.login();
  }

  protected byte[] getBarContent(final URL xpdlURL) {
    try {
      return BusinessArchiveFactory.getBusinessArchiveContent(xpdlURL, (Class[]) null);
    } catch (final IOException e) {
      e.printStackTrace();
      fail("unable to build businessArchive");
      return null;
    }
  }

  protected byte[] getBarContent(final byte[] xpdlContent) {
    try {
      return BusinessArchiveFactory.getBusinessArchiveContent(xpdlContent, (Class[]) null);
    } catch (final IOException e) {
      e.printStackTrace();
      fail("unable to build businessArchive: " + Misc.getStackTraceFrom(e));
      return null;
    }
  }

  protected byte[] getBarContent(final URL xpdlURL, final Class< ? >... classes) {
    try {
      return BusinessArchiveFactory.getBusinessArchiveContent(xpdlURL, classes);
    } catch (final IOException e) {
      e.printStackTrace();
      fail("unable to build businessArchive: " + Misc.getStackTraceFrom(e));
      return null;
    }
  }

  protected byte[] getBarContent(final byte[] xpdlContent, final Class< ? >... classes) {
    try {
      return BusinessArchiveFactory.getBusinessArchiveContent(xpdlContent, classes);
    } catch (final IOException e) {
      e.printStackTrace();
      fail("unable to build businessArchive: " + Misc.getStackTraceFrom(e));
      return null;
    }
  }

  protected Set<byte[]> getClasses(final Class< ? >... classes) {
    try {
      return ClassDataTool.getClasses(classes);
    } catch (final Exception e) {
      e.printStackTrace();
      fail("unable to build businessArchive: " + Misc.getStackTraceFrom(e));
      return null;
    }
  }

}
