package org.ow2.bonita.activity.instantiation;

import java.net.URL;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.activity.instantiation.instantiator.BadJoinNActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.BadVarActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.HookActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.JoinNumberConditionFail;
import org.ow2.bonita.activity.instantiation.instantiator.JoinNumberConditionPass;
import org.ow2.bonita.activity.instantiation.instantiator.NullActivityInstantiator;
import org.ow2.bonita.activity.instantiation.instantiator.TestActivityInstantiator;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.ExpressionEvaluationException;
import org.ow2.bonita.facade.exception.MultiInstantiatorInvocationException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;

public class MultiInstantiationTest extends APITestCase {

  public void testMultiInstantiation() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("multiInstantiation.xpdl");
    String processId = "multiInstantiation";

    Map<String, ProcessDefinition> processMap = getManagementAPI().deployBar(getBarContent(xpdlUrl,
      TestActivityInstantiator.class));
    ProcessDefinition process = processMap.get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    checkExecutedManyTimes(instanceUUID, new String[]{"r1"}, 2);
    
    Set<ActivityInstance<ActivityBody>> actIntances = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
    
    Set<Object> varValues = new HashSet<Object>();
    
    for (ActivityInstance<ActivityBody> activityInstance : actIntances) {
      Map<String, Object> variables = activityInstance.getLastKnownVariableValues();
      assertTrue(variables.containsKey("testVar"));
      varValues.add(variables.get("testVar"));
    }
    
    Set<Object> expectedValues = new HashSet<Object>();
    expectedValues.add("val1");
    expectedValues.add("val2");
    
    assertEquals(expectedValues, varValues);
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  
  public void testDeployWrongInstantiator() {
    //Test deployment with multiInstantiatorClass class not found
    URL xpdlUrl = this.getClass().getResource("incorrectMultiInstantiation.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Deployment shouldn't passed");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a MultiInstantiator class name"));
      assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a variable id"));

    }
  }
  
  public void testTransitionsConditions() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("transitionConditionMultiInstantiation.xpdl");
    Map<String, ProcessDefinition> processMap = getManagementAPI().deployBar(getBarContent(
      xpdlUrl, JoinNumberConditionPass.class, JoinNumberConditionFail.class));
    //Test join number with transition condition on activity variable
    // this is not allowed
    String processId = "transitionCondition";
    ProcessDefinitionUUID processUUID = processMap.get(processId).getUUID();
    try {
      getRuntimeAPI().instantiateProcess(processUUID);
      fail("Unreachable statement !");
    } catch (ExpressionEvaluationException e) {
      String expected = "conditionVar";
      assertTrue("Expected: " + expected + ", result: " + e.getMessage(), e.getMessage().contains(expected));
    }
    Set<ProcessInstance> processInstances = getQueryRuntimeAPI().getProcessInstances(processUUID); 
    assertTrue(processInstances == null || processInstances.isEmpty());
    
    PackageDefinitionUUID packageUUID = processMap.get(processId).getPackageDefinitionUUID();
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testHookMultiInstantiation() throws BonitaException {
    //Test hook execution with multi instantiation
    URL xpdlUrl = this.getClass().getResource("hookMultiInstantiation.xpdl");
    String processId = "hookMultiInstantiation";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
      HookActivityInstantiator.class, IncrInstanceVariableHook.class)).get(processId);
    
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(process.getUUID());
    checkExecutedManyTimes(instanceUUID, new String[]{"r1"}, 3);
    
    Set<ActivityInstance<ActivityBody>> actIntances = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
    
    Set<Object> varValues = new HashSet<Object>();
    Set<Object> expectedValues = new HashSet<Object>();
    
    for (ActivityInstance<ActivityBody> activity : actIntances) {
      Map<String, Object> variables = activity.getLastKnownVariableValues();
      assertTrue(variables.containsKey("counter"));
      varValues.add(variables.get("counter"));
    }
    
    expectedValues.add(4L);
    expectedValues.add(5L);
    expectedValues.add(6L);
    
    assertEquals(expectedValues, varValues);
    
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  /*
   * Multiinstantiation element with a variable not declared in the XPDL 
   */
  public void testInvalidVarName() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    URL xpdlUrl = this.getClass().getResource("invalidVarNameMultiInstantiation.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, BadVarActivityInstantiator.class));
      fail("exception expected");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage().contains("MultiInstantiation variable testVar must be a local variable of activity r1"));
    }
  }
  
  /*
   * Multiinstantiation element without variable 
   */
  public void testNoVariable() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    URL xpdlUrl = this.getClass().getResource("noVarMultiInstantiation.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, BadVarActivityInstantiator.class));
      fail("exception expected");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage().contains("MultiInstantiation needs to specify a variable id (in a nested Variable element)"));
    }
  }
  
  public void testFailInstantiator() throws BonitaException {
    //Test multiInstantiatorClass which throws an exception
    URL xpdlUrl = this.getClass().getResource("failMultiInstantiation.xpdl");
    String processId = "multiInstantiation";
    Map<String, ProcessDefinition> processMap = getManagementAPI().deployBar(getBarContent(xpdlUrl,
      NullActivityInstantiator.class, BadJoinNActivityInstantiator.class));
    ProcessDefinition process = processMap.get(processId);
    try {
      getRuntimeAPI().instantiateProcess(process.getUUID());
      fail("Failing multiInstantiatorClass should throw an exception");
    } catch (MultiInstantiatorInvocationException e) {
      assertTrue(causeContains("MultiInstantiator execution returned null in activity r1", e));
    }
    getRuntimeAPI().deleteAllProcessInstances(process.getUUID());

    processId = "invalidJoinNumber";
    process = processMap.get(processId);
    try {
      getRuntimeAPI().instantiateProcess(process.getUUID());
      fail("Process instantiation shouldn't passed because of invalid join number");
    } catch (MultiInstantiatorInvocationException e) {
      assertTrue(causeContains("The join number must be greater than 0", e));
    }
    
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();  
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  // multi instantiation element referring to global variable
  public void testGlobalVariableMultiInstantiation() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("globalVarMultiInstantiation.xpdl");

    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl,
          TestActivityInstantiator.class));
      fail("Deployment should fail");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage().contains("MultiInstantiation variable testVar must be a local variable of activity r1"));
    }
  }
  
  

}
