package org.ow2.bonita.activity.subflow;

import java.net.URL;
import java.util.Collection;
import java.util.Map;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;

public class SubflowTest extends APITestCase {

  /**
   * Automatic process that creates a subflow with only Automatic activities.
   */
  public void testAutoSubFlow() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("subflowAuto.xpdl");
    String processId = "main";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "a2", "a3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /**
  * Automatic process that creates a subflow with a Manual activity.
  */
  public void testSubFlowWithTask() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("subflow.xpdl");
    String processId = "main";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"a1"});

    Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());
    ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();
    ProcessInstance processInstance = getQueryRuntimeAPI().getProcessInstance(taskActivity.getProcessInstanceUUID());
    assertEquals(instanceUUID, processInstance.getParentInstanceUUID());
    TaskUUID taskUUID = taskActivity.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "a2", "a3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * Main process creates a subflow instance, which creates a new subflow.
   */
  public void testNestedSubFlowWithTask() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("nestedSubflow.xpdl");
    String processId = "main";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"a1"});

    Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());
    TaskUUID taskUUID = taskActivities.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "a2", "a3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testSubFlowWithTaskAndParams() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("subflowParams.xpdl");
    String processId = "main";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"a1"});

    assertEquals("test", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param1"));
    assertEquals("initial", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param2"));
    assertEquals("initial", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param3"));

    Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());
    ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();

    ProcessInstanceUUID taskInstanceUUID = taskActivity.getProcessInstanceUUID();
    assertNotSame(instanceUUID, taskInstanceUUID);

    assertEquals("test", getQueryRuntimeAPI().getVariable(taskActivity.getUUID(), "inParam"));
    assertEquals("initial", getQueryRuntimeAPI().getVariable(taskActivity.getUUID(), "inoutParam"));
    // out params are initialized with a null value
    // Xpdl 1.0 spec specifies in 7.1.2.1 that out parameters are initialized to zero
    // (empty string for strings, zero for each component of complex data)
    // It 's easier to set the variable to null, but maybe this needs to be changed.
    assertEquals(null, getQueryRuntimeAPI().getVariable(taskActivity.getUUID(), "outParam"));

    TaskUUID taskUUID = taskActivity.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().setProcessInstanceVariable(taskInstanceUUID, "outParam", "changed");
    getRuntimeAPI().setProcessInstanceVariable(taskInstanceUUID, "inoutParam", "changed");
    getRuntimeAPI().setProcessInstanceVariable(taskInstanceUUID, "inParam", "changed");
    getRuntimeAPI().finishTask(taskUUID, true);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "a2", "a3"});

    assertEquals("test", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param1"));
    assertEquals("changed", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param2"));
    assertEquals("changed", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "param3"));

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * ProEd use a subflow process defined in an external file.
   * This process needs to be deployed separately
   */
  public void testAutoSubFlowProEd() throws BonitaException {
    URL xpdlUrlSubflow = this.getClass().getResource("automaticProcess.xpdl");
    PackageDefinitionUUID subflowPackageUUID =
      getManagementAPI().deployBar(getBarContent(xpdlUrlSubflow)).values().iterator().next().getPackageDefinitionUUID();

    URL xpdlUrl = this.getClass().getResource("subflow_proed.xpdl");
    String processId = "subflow_proed";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "a2", "a3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().undeploy(subflowPackageUUID);
    
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().deletePackage(subflowPackageUUID);
  }

  /**
   * Checks that deployment fails if the subFlow process is not
   * already deployed on the repository.
   */
  public void testUnknownSubFlowProcess() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("subflow_proed.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected exception : subflow process not found");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains("Unable to find process"));
    }
  }

  /**
   * Checks that it is not possible to undeploy a package
   * that is used by another deployed package.
   */
  public void testUndeploySubProcess() throws BonitaException {
    URL xpdlUrlSubflow = this.getClass().getResource("automaticProcess.xpdl");
    PackageDefinitionUUID subflowPackageUUID =
      getManagementAPI().deployBar(getBarContent(xpdlUrlSubflow)).values().iterator().next().getPackageDefinitionUUID();
    URL xpdlUrl = this.getClass().getResource("subflow_proed.xpdl");
    PackageDefinitionUUID packageUUID =
      getManagementAPI().deployBar(getBarContent(xpdlUrl)).values().iterator().next().getPackageDefinitionUUID();

    try {
      getManagementAPI().undeploy(subflowPackageUUID);
      fail("Expected exception");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains(packageUUID.toString()));
    }

    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().undeploy(subflowPackageUUID);
    
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().deletePackage(subflowPackageUUID);
    
  }

  /**
   * ProEd use a subflow process defined in an external file.
   * This process needs to be deployed separately
   */
  public void testSubFlowProEdVariables() throws BonitaException {
    URL xpdlUrlSubflow = this.getClass().getResource("subflow_proed_variables.xpdl");
    PackageDefinitionUUID subflowPackageUUID =
      getManagementAPI().deployBar(getBarContent(xpdlUrlSubflow)).values().iterator().next().getPackageDefinitionUUID();
    URL xpdlUrl = this.getClass().getResource("main_variables_proed.xpdl");
    String processId = "main_variables_proed";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkStopped(instanceUUID, new String[]{"a1"});

    Collection<ActivityInstance< TaskInstance >> tasks = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, tasks.size());
    ActivityInstance< TaskInstance > taskInstance = tasks.iterator().next();
    ProcessInstance processInstance = getQueryRuntimeAPI().getProcessInstance(taskInstance.getProcessInstanceUUID());
    assertEquals(instanceUUID, processInstance.getParentInstanceUUID());
    TaskUUID taskUUID = taskInstance.getBody().getUUID();

    Map<String, Object> variables =
      getQueryRuntimeAPI().getActivityInstanceVariables(taskInstance.getUUID());
    assertNotNull(variables);
    assertTrue(variables.containsKey("activityVar1"));
    assertFalse(variables.containsKey("processVar1"));
    assertFalse(variables.containsKey("subflow_var"));
    assertFalse(variables.containsKey("parent_process_var"));

    variables = getQueryRuntimeAPI().getProcessInstanceVariables(taskInstance.getProcessInstanceUUID());
    assertFalse(variables.containsKey("activityVar1"));
    assertTrue(variables.containsKey("processVar1"));
    assertTrue(variables.containsKey("subflow_var"));
    assertFalse(variables.containsKey("parent_process_var"));
    
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);

    checkExecutedOnce(instanceUUID, new String[]{"a1", "sf"});

    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "sf");
    assertEquals(1, acts.size());
    ActivityInstanceUUID sfUUID = acts.iterator().next().getUUID();
    variables = getQueryRuntimeAPI().getActivityInstanceVariables(sfUUID);

    assertNotNull(variables);
    assertFalse(variables.containsKey("activityVar1"));
    assertTrue(variables.containsKey("subflow_var"));


    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().undeploy(subflowPackageUUID);
    
    getManagementAPI().deletePackage(packageUUID);
    getManagementAPI().deletePackage(subflowPackageUUID);
  }
}
