/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.deadline;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;

/**
 * @author Pascal Verdage
 */
public class DeadlineFunctionalTest extends APITestCase {

  private static final long TIMEOUT = 3 * 60 * 1000; // 3 minutes
  /** */
  protected long waitForLongVariableToBeSet(final ProcessInstanceUUID instanceUUID,
      final String variable, final long timeout) {
    long result = Long.MAX_VALUE;
    final long rate = 500;
    for (long i = 0; i < timeout / rate; i++) {
      try {
        Thread.sleep(rate);
        final Object object = getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, variable);
        if (object != null && object instanceof String) {
          result = Long.parseLong((String) object);
          if (result != 0) {
            break;
          }
        }
      } catch (final Exception e) {
        e.printStackTrace();
        fail(e.getMessage());
      }
    }
    return result;
  }

  protected long instantiateAndWaitForLongVariableToBeSet(final ProcessDefinition process,
      final String variable) throws ProcessNotFoundException,
      DeploymentException, PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    return instantiateAndWaitForLongVariableToBeSet(process, variable, TIMEOUT);
  }

  protected long instantiateAndWaitForLongVariableToBeSet(final ProcessDefinition process,
      final String variable, final long timeout) throws ProcessNotFoundException,
      DeploymentException, PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    long time;
    try {
      final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      time = waitForLongVariableToBeSet(instanceUUID, variable, timeout);
    } finally {
      getRuntimeAPI().deleteAllProcessInstances(processUUID);
      getManagementAPI().clearHistory();
      getManagementAPI().undeploy(packageUUID);
    }
    return time;
  }

  /** test the execution of a transactional hook after a fixed duration */
  public void testTxHookAfterDuration() throws BonitaException {
    final Class< ? > hook = SetTimeTxHook.class;
    final String processId = "main";
    final String packageId = "HookAfterDuration";
    final String document = new DeadlineDocument(packageId, "Test of a hook taken after a duration")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, "1000", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    final ProcessDefinition process = 
      getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    final long startTime = (System.currentTimeMillis() / 1000) * 1000;
    final long time = instantiateAndWaitForLongVariableToBeSet(process, "time");

    assertTrue("execution duration: " + (time - startTime), time - startTime >= 1000);
    getManagementAPI().deletePackage(process.getPackageDefinitionUUID());
  }

  /** test the execution of a non-transactional hook after a fixed duration */
  public void testHookAfterDuration() throws BonitaException {
    final Class< ? > hook = SetTimeHook.class;
    final String packageId = "HookAfterDuration";
    final String processId = "main";
    final String document = new DeadlineDocument(packageId, "Test of a transactional hook taken after a duration")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, "1000", hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    final long startTime = (System.currentTimeMillis() / 1000) * 1000;
    final long time = instantiateAndWaitForLongVariableToBeSet(process, "time");

    assertTrue("execution duration: " + (time - startTime), time - startTime >= 1000);
    getManagementAPI().deletePackage(process.getPackageDefinitionUUID());
  }

  /** test the execution of a transactional hook at a fixed date */
  public void testTxHookAtDate() throws BonitaException {
    final long startTime = (System.currentTimeMillis() / 1000) * 1000;
    final Class< ? > hook = SetTimeTxHook.class;
    final String processId = "main";
    final String packageId = "TxHookAtDate";
    final String deadlineCondition = DeadlineDocument.createDeadlineDateFromDelay(1000);
    final String document = new DeadlineDocument(packageId, "Test of a transactional hook taken at a fixed date")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, deadlineCondition, hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    final ProcessDefinition process = 
      getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    final long time = instantiateAndWaitForLongVariableToBeSet(process, "time");

    assertTrue("execution duration: " + (time - startTime), time - startTime >= 1000);
    getManagementAPI().deletePackage(process.getPackageDefinitionUUID());
  }

  /** test the execution of a non-transactional hook at a fixed date */
  public void testHookAtDate() throws BonitaException {
    final long startTime = (System.currentTimeMillis() / 1000) * 1000;
    final Class< ? > hook = SetTimeHook.class;
    final String processId = "main";
    final String packageId = "HookAtDate";
    final String deadlineCondition = DeadlineDocument.createDeadlineDateFromDelay(1000);
    final String document = new DeadlineDocument(packageId, "Test of a hook taken at a fixed date")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, deadlineCondition, hook.getName())
      .endActivity().endActivities().endProcess().endDocument().getDocument();

    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(document.getBytes(), hook)).get(processId);
    final long time = instantiateAndWaitForLongVariableToBeSet(process, "time");

    assertTrue("execution duration: " + (time - startTime), time - startTime >= 1000);
    getManagementAPI().deletePackage(process.getPackageDefinitionUUID());
  }

  /** test the cancellation of a deadline when the instance is deleted */
  public void testDeadlinesCanceledByDeleteInstance() throws BonitaException {
    final String processId = "main";
    final String packageId = "CancelledDeadlines";
    final DeadlineDocument document = new DeadlineDocument(packageId, "Test of cancelled deadlines")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, "1000", NullHook.class.getName())
      .addDeadline(null, "10000", SetTimeTxHook.class.getName())
      .addDeadline(null, "10000", SetTimeHook.class.getName())
      .endActivity().endActivities().endProcess().endDocument();

    final ProcessDefinition process = getManagementAPI().deployBar(
        getBarContent(document.getDocument().getBytes(), NullHook.class, SetTimeHook.class, SetTimeTxHook.class)).get(processId);
    final long time = instantiateAndWaitForLongVariableToBeSet(process, "time", 2000);
    assertEquals(0, time);
    getManagementAPI().deletePackage(process.getPackageDefinitionUUID());
  }

  /** test the cancellation of a deadline when the activity is executed */
  public void testDeadlinesCanceledByActivity() throws BonitaException {
    final Class< ? > hook = SetTimeHook.class;
    final String processId = "main";
    final String packageId = "CancelledDeadlines";
    final DeadlineDocument document = new DeadlineDocument(packageId, "Test of cancelled deadlines")
      .startProcess(processId, "main")
      .addDataField("time", "0")
      .addDefaultParticipant()
      .startActivity("a", "a")
      .addDeadline(null, "10000", hook.getName())
      .addDeadline(null, "10000", hook.getName())
      .endActivity().startActivity("b", "b")
      .endActivity().endActivities()
      .addTransition("a_b", "a_b", "a", "b")
      .endProcess().endDocument();

    final ProcessDefinition process = getManagementAPI().deployBar(
        getBarContent(document.getDocument().getBytes(), hook)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();
    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    long time = Long.MAX_VALUE;
    try {
      checkStopped(instanceUUID, new String[]{});
      executeTask(instanceUUID, "a");
      time = waitForLongVariableToBeSet(instanceUUID, "time", 12000);
    } finally {
      getRuntimeAPI().deleteAllProcessInstances(processUUID);
      getManagementAPI().clearHistory();
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    }
    assertEquals(0, time);
  }

}
