package org.ow2.bonita.facade;

import java.net.URL;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.novabpm.util.SimpleCallbackHandler;

public class RuntimeAPITest extends APITestCase {

 
  /*
   *   test instantiateProcess(String processDefinitionUUID) from RuntimeAPI
   *
   *  Test:
   *   -instantiation + start execution
   *   -returned instance has STARTED state
   *  test :
   *     throws ProcessNotFoundException if the process has not been found
   *     (ex. : ""  random processDefinitionUUID) ,
   *     throws BonitaInternalException if an exception occurs (ex. if processDefinitionUUID == null)
   */
  public void testInstantiateProcess() throws BonitaException {
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      getRuntimeAPI().instantiateProcess(null);
      fail("Check null parameter");
    } catch (IllegalArgumentException be) {
      // OK
    }
    try {
      getRuntimeAPI().instantiateProcess(IdFactory.getNewProcessUUID());
      fail("Check non existent process");
    } catch (ProcessNotFoundException be) {
      //OK
    }
    try {
      final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      // instantiateProcess calls start so the instance is started.
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());

      getRuntimeAPI().deleteProcessInstance(instanceUUID);
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    } catch (BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      be.printStackTrace();
      fail("Exception thrown!");
    }
  }

  /*
   *  Test instantiateProcess(String processDefinitionUUID, Map<String Object>variables) from RuntimeAPI
   *    check the map of variables has been added to the created instance.
   */
  public void testInstantiateProcessWithMapVariables() throws BonitaException {
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      Map<String, Object> expectedVarMap = new HashMap<String, Object>();
      expectedVarMap.put("str1", "toto");
      final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID, expectedVarMap);
      // instantiateProcess calls start so the instance is started.
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());

      Map<String, Object> currentVarMap = getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
      //check the instance get the expected variable map.
      assertEquals(expectedVarMap, currentVarMap);

      getRuntimeAPI().deleteProcessInstance(instanceUUID);
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    } catch (BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      be.printStackTrace();
      fail("Exception thrown!");
    }
  }

  /*
   *  test deleteInstance & deleteInstances
   */
  public void testDeleteInstance() throws BonitaException {
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    try {
      //test deleteInstance()
      final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      Collection<ActivityInstance< TaskInstance >> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
      assertTrue("Check task is not empty", tasks.size() == 1);
      getRuntimeAPI().deleteProcessInstance(instanceUUID);
      try {
        //getInstance call the journal !!
        getQueryRuntimeAPI().getProcessInstance(instanceUUID);
        fail("Check fail getting instance into the journal with getInstance()");
      } catch (InstanceNotFoundException ie) {
        assertTrue("Check exception has well been raised !", ie.getInstanceUUID().equals(instanceUUID));
      }

      //test deleteInstances()
      ProcessInstanceUUID instanceUUID1 = getRuntimeAPI().instantiateProcess(processUUID);
      ProcessInstanceUUID instanceUUID2 = getRuntimeAPI().instantiateProcess(processUUID);
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID1).getInstanceState());
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID2).getInstanceState());

      tasks = getQueryRuntimeAPI().getTaskList(TaskState.READY);
      assertEquals("Check task is not empty", 2, tasks.size());

      getRuntimeAPI().deleteAllProcessInstances(processUUID);

      Set<ProcessInstance> processInstances = getQueryRuntimeAPI().getProcessInstances(processUUID);
      assertTrue("No more process instance into the journal", processInstances.size() == 0);
      getManagementAPI().undeploy(packageUUID);
      getManagementAPI().deletePackage(packageUUID);
    } catch (BonitaException be) {
      if (be.getCause() instanceof ProcessNotFoundException) {
        throw new RuntimeException("Missing resource!", be);
      }
      be.printStackTrace();
      fail("Exception thrown!");
    }
  }

  /*
   *  Test startTask(String taskUUID) from the runtimeAPI.
   *
   * test initial state need to be READY.
   * change state to EXECUTING.
   * tests following exceptions are raised:
   *  @throws TaskNotFoundException if the task has not been found.
   *  @throws IllegalTaskStateException if the state of the task has not READY state.
   *  @throws BonitaInternalException if an exception occurs (ex. : if taskUUID == null)
   */
  public void testStartTask() throws BonitaException {
    ProcessInstanceUUID instanceUUID = null;
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    } catch (BonitaException e) {
      e.printStackTrace();
      fail("Ouch! How can it be?! Should never reach this statement!");
    }
    try {
      getRuntimeAPI().startTask(null, true);
      fail("Check null argument");
    } catch (IllegalArgumentException be) {
      // ok
    }
    try {
      getRuntimeAPI().startTask(IdFactory.getNewTaskUUID(), true);
      fail("Check non existent task");
    } catch (TaskNotFoundException be) {
      //OK
    }
    checkStopped(instanceUUID, new String[]{});
    Collection<ActivityInstance< TaskInstance >> activitiesToDo = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, activitiesToDo.size());
    ActivityInstance< TaskInstance > taskActivity = activitiesToDo.iterator().next();
    assertEquals(TaskState.READY, getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    try {
      getRuntimeAPI().startTask(taskActivity.getBody().getUUID(), true);
    } catch (BonitaException e) {
      e.printStackTrace();
      fail("Ouch! How can it be?! Should never reach this statement!");
    }
    assertEquals(TaskState.EXECUTING, getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test terminateTask(String taskUUID) from the runtimeAPI.
   *
   * Test initial state required is : EXECUTING.
   * Test if finish is successful state become FINISHED.
   * Test
   *    - @throws TaskNotFoundException if the task has not been found.
   *    - @throws IllegalTaskStateException if the state of the task has not EXECUTING state.
   *    - @throws BonitaInternalException if an exception occurs (for ex. with illegal taskUUID parameter).
   *
   */
  public void testFinishTask() throws BonitaException {
    ProcessInstanceUUID instanceUUID = null;
    String processId = "usertest1";
    URL xpdlUrl = this.getClass().getResource("usertest1_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    } catch (BonitaException e) {
      fail("Ouch! How can it be?! Should never reach this statement!");
    }
    try {
      getRuntimeAPI().finishTask(null, true);
      fail("Check null argument");
    } catch (IllegalArgumentException be) {
      // ok
    }
    try {
      getRuntimeAPI().finishTask(IdFactory.getNewTaskUUID(), true);
      fail("Check non existent instance");
    } catch (TaskNotFoundException be) {
      //OK
    }
    checkStopped(instanceUUID, new String[]{});
    Collection<ActivityInstance< TaskInstance >> activitiesToDo = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, activitiesToDo.size());
    ActivityInstance< TaskInstance > taskActivity = activitiesToDo.iterator().next();
    try {
      getRuntimeAPI().startTask(taskActivity.getBody().getUUID(), true);
    } catch (BonitaException e) {
      e.printStackTrace();
      fail("Ouch! How can it be?! Should never reach this statement!");
    }
    activitiesToDo = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.EXECUTING);
    assertEquals(1, activitiesToDo.size());
    taskActivity = activitiesToDo.iterator().next();
    assertEquals(TaskState.EXECUTING, getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());

    //Check IllegalTaskStateException (suspend the task and try to finish)
    try {
      getRuntimeAPI().suspendTask(taskActivity.getBody().getUUID(), true);
      getRuntimeAPI().finishTask(taskActivity.getBody().getUUID(), true);
    } catch (IllegalTaskStateException be) {
      //OK
    }

    try {
      getRuntimeAPI().resumeTask(taskActivity.getBody().getUUID(), false);
      Collection<ActivityInstance<TaskInstance>> resumedTasks = getQueryRuntimeAPI().getTaskList("admin", TaskState.EXECUTING);
      assertNotNull(resumedTasks);
      assertEquals(1, resumedTasks.size());
      getRuntimeAPI().finishTask(taskActivity.getBody().getUUID(), true);
    } catch (BonitaException e) {
      e.printStackTrace();
      fail("Ouch! How can it be?! Should never reach this statement!");
    }

    assertEquals(TaskState.FINISHED, getQueryRuntimeAPI().getTask(taskActivity.getBody().getUUID()).getBody().getState());
    Collection<ActivityInstance<TaskInstance>> finishedTasks = getQueryRuntimeAPI().getTaskList("admin", TaskState.FINISHED);
    assertNotNull(finishedTasks);
    assertEquals(1, finishedTasks.size());
    ActivityInstance<TaskInstance> finishedTask = finishedTasks.iterator().next();  
    assertEquals(TaskState.FINISHED, finishedTask.getBody().getState());
    assertEquals(taskActivity.getActivityId(), finishedTask.getActivityId());
    
    // Try to finish a finished task
    try {
      getRuntimeAPI().finishTask(finishedTask.getBody().getUUID(), true);
      fail("expecting exception");
    } catch (IllegalTaskStateException e) {
      // ok
    }
    // Try to start a finished task
    try {
      getRuntimeAPI().startTask(finishedTask.getBody().getUUID(), true);
      fail("expecting exception");
    } catch (IllegalTaskStateException e) {
      // ok
    }
    
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   *  test exceptions and the documented case from javadoc
   *  (if the instance or activity var does not exist it is created)
   *  for methods (of RuntimeAPI) :
   *     - setInstanceVariable
   *     - setActivityVariable
   *
   *     TODO (not tested):
   *       - test the AmbiguousActivityException raised if multiple iterations exist !!!!!
   *       - test ActivityNotFoundException if setting activity variables when activty
   *          already finished
   *       - test InstanceNotFoundException if setting instance variables when instance
   *          already finished
   *      Note:   need probably to introduce a new exception for that !!!!!
   */
  public void testSetVariables() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    String activity1 = "act1";
    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    assertEquals("act1", activityInst.getActivityId());

    //test InstanceNotFoundException
    try {
      getRuntimeAPI().setProcessInstanceVariable(badInstanceUUID, "string_process", "titi");
      fail("Check non existent Instance");
    } catch (InstanceNotFoundException be) {
      //OK
      assertEquals(badInstanceUUID, be.getInstanceUUID());
    }

    //test setActivityVariable with activityUUID
    try {
      getRuntimeAPI().setActivityInstanceVariable(badActivityUUID, "string_activity", "tutu");
      fail("Check non existent Activity Instance");
    } catch (ActivityNotFoundException be) {
      //OK
      assertEquals(badActivityUUID, be.getActivityUUID());
    }

    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});

    //get info with "act2" activity
    String activity2 = "act2";
    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity2);
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();
    assertEquals(activity2, activityInst.getActivityId());


    //**** testing setVariable() ****
    //with valid parameter to set process instance variable
    getRuntimeAPI().setVariable(activityUUID, "string_process", "new_titi");
    assertEquals("new_titi", getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_process"));

    //with valid parameter to set activity instance variable
    getRuntimeAPI().setVariable(activityUUID, "string_activity", "new_titi");
    assertEquals("new_titi",getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_activity"));


    //with invalid parameters to test exceptions:
    //InstanceNotFoundException
    //TODO
    try {
      getRuntimeAPI().setVariable(badActivityUUID, "string_process", "titi");
      fail("Check non existent Instance");
    } catch (ActivityNotFoundException be) {
      //OK
      assertEquals(badActivityUUID, be.getActivityUUID());
    }

    //VariableNotFoundException
    try {
      getRuntimeAPI().setVariable(activityUUID, "string_process_bidon", "titi");
      fail("Check non existent Instance");
    } catch (VariableNotFoundException be) {
      //OK
      assertEquals("string_process_bidon", be.getVariableId());
    }

    // start & terminate activity2 task
    executeTask(instanceUUID, activity2);

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }



  /*
   * Testing getActivityvariables() without cycle/iteration
   * There's 2 methods for getActivityVariables()
   * 1) Map<String, Object> getActivityVariables(ProcessInstanceUUID instanceUUID,String activityId)
   *  throws: ActivityNotFoundException,
   *           AmbiguousActivityException,
   *           InstanceNotFoundException;
   *  2) Map<String, Object> getActivityVariables(ActivityInstanceUUID activityUUID)
   *   throws ActivityNotFoundException;
   *
   *  Check also if trying with wrong iterationId an ActivityNotFoundException is raised
   *  (specific exception with the iteration id as parameter).
   */
  public void testGetActivityVariables() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    String activity1 = "act1";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    //ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();

    //in act1 there's only process variables so the method returns an empty list
    Map<String, Object> variables = getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    assertTrue(variables.isEmpty());

    //Check raised exceptions with getActivityVariables(ActivityInstanceUUID activityUUID)
    // 1- ActivityNotFoundException
    try {

      getQueryRuntimeAPI().getActivityInstanceVariables(badActivityUUID);
      fail("Check non existent activity");
    } catch (ActivityNotFoundException be) {
      //OK
    }

    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});

    //execution is poniting to act2 which contains process and activity variables
    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act2");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();

    //when using parameters: activityUUID
    variables = getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_activity"));
    assertTrue(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node


    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act3");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();
    variables = getQueryRuntimeAPI().getActivityInstanceVariables(activityUUID);
    assertTrue(variables.isEmpty());

    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetProcessVariables() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    String activity1 = "act1";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //in act1 there's only process variables
    Map<String, Object> variables = getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));

    //Check RAISED EXCEPTIONS
    // InstanceNotFoundException
    try {
      getQueryRuntimeAPI().getProcessInstanceVariables(badInstanceUUID);
      fail("Check non existent instance");
    } catch (InstanceNotFoundException be) {
      //OK
    }

    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});

    // Check variables within execution pointing to act2 node
    variables = getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));
    assertFalse(variables.containsKey("string_activity"));
    assertFalse(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");

    // check variables within execution pointing to act3 node
    variables = getQueryRuntimeAPI().getProcessInstanceVariables(instanceUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));
    assertFalse(variables.containsKey("string_activity"));
    assertFalse(variables.containsKey("enum_activity"));

    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetVariables() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    String activity1 = "act1";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();
    
    //in act1 there's only process variables
    Map<String, Object> variables = getQueryRuntimeAPI().getVariables(activityUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));

    //Check RAISED EXCEPTIONS
    // ActivityNotFoundException
    try {
      getQueryRuntimeAPI().getVariables(badActivityUUID);
      fail("Check non existent activity");
    } catch (ActivityNotFoundException be) {
      //OK
    }


    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});

    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act2");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();
    
    variables = getQueryRuntimeAPI().getVariables(activityUUID);
    assertFalse(variables.isEmpty());
    assertEquals(4, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));
    assertTrue(variables.containsKey("string_activity"));
    assertTrue(variables.containsKey("enum_activity"));

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node
    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act3");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    activityUUID = activityInst.getUUID();
    
    variables = getQueryRuntimeAPI().getVariables(activityUUID);
    assertFalse(variables.isEmpty());
    assertEquals(2, variables.size());
    assertTrue(variables.containsKey("string_process"));
    assertTrue(variables.containsKey("enum_process"));
    assertFalse(variables.containsKey("string_activity"));
    assertFalse(variables.containsKey("enum_activity"));

    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetActivityInstance() throws BonitaException {
    String processId = "usertest2";
    URL xpdlUrl = this.getClass().getResource("usertest2_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    try {
      final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      //generating bad UUID to test exceptions
      ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();
      
      Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a");
      assertEquals(1, acts.size());
      ActivityInstanceUUID activityUUID = acts.iterator().next().getUUID();
      
      ActivityInstance<ActivityBody> activityInst = getQueryRuntimeAPI().getActivityInstance(activityUUID);
      assertNotNull(activityInst);
      assertEquals("a", activityInst.getActivityId());
      //check ActivityNotFoundException
      try {
        activityInst = getQueryRuntimeAPI().getActivityInstance(badActivityUUID);
      } catch (ActivityNotFoundException ae) {
        assertEquals(badActivityUUID, ae.getActivityUUID());
      }
    } catch (ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Testing getActivityVariable() without cycle/iteration
   *
   *  getActivityVariable(ActivityInstanceUUID activityUUID,String variableId)
   *     throws : ActivityNotFoundException,
   *              VariableNotFoundException;
   */
  public void testGetActivityVariable() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    String activity1 = "act1";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    //ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();
    ActivityInstanceUUID badActivityUUID = IdFactory.getNewActivityUUID();

    // Check variables within execution pointing to act1 node
    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, activity1);
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();

    //Check getActivityVariable(ActivityInstanceUUID activityUUID,String variableId) method
    //process var
    try {
      getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_process");
      fail("string_process is not local to act1 !");
    } catch (VariableNotFoundException ve) {
      //OK
      assertEquals("string_process", ve.getVariableId());
    }

    //activity var does not exist
    //VariableNotFoundException
    try {
      getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "string_activity");
      fail("Check the activity var does not exist in act1");
    } catch (VariableNotFoundException ve) {
      //OK
      assertEquals("string_activity", ve.getVariableId());
    }
    //ActivityNotFoundException
    try {
      getQueryRuntimeAPI().getActivityInstanceVariable(badActivityUUID, "string_process");
      fail("Check non existent badActivityUUID");
    } catch (ActivityNotFoundException ae) {
      //OK
      assertEquals(badActivityUUID, ae.getActivityUUID());
    }

    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});
    // activityInst is unused
    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act2");

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node

    //    activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act3");
    // activityInst is unused
    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testGetProcessVariable() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("variables.xpdl");
    String processId = "variables";
    String activity1 = "act1";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    //generating bad UUID to test exceptions
    ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //Check getActivityVariable(ActivityInstanceUUID activityUUID,String variableId) method
    //process var
    String varStringProcess = (String)getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_process");
    assertTrue(varStringProcess.equals("initial string value"));

    //No activity var exist
    //VariableNotFoundException
    try {
      getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "string_activity");
      fail("Check the activity var does not exist in act1");
    } catch (VariableNotFoundException ve) {
      //OK
      assertEquals("string_activity", ve.getVariableId());
    }
    //InstanceNotFoundException
    try {
      varStringProcess = (String)getQueryRuntimeAPI().getProcessInstanceVariable(badInstanceUUID, "string_process");
      fail("Check non existent badActivityUUID");
    } catch (InstanceNotFoundException ae) {
      //OK
      assertEquals(badInstanceUUID, ae.getInstanceUUID());
    }

    // start & terminate "act1" task
    executeTask(instanceUUID, activity1);
    checkStopped(instanceUUID, new String[]{activity1});
    // activityInst is unused
    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act2");

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node

    // activityInst = getQueryRuntimeAPI().getActivityInstance(instanceUUID, "act3");
    // activityInst is unused
    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test Collection<TaskInstance> getTaskList(TaskState taskState);
   * from RuntimeReadOnlyAPI
   */

  //Takes into account
  //    . the states (4 states available)
  //    . the 3 types of methods depending on the parameters:
  //                . TaskState taskState,
  //                . ProcessInstanceUUID instanceUUID,
  //                . java.lang.String userId
  // Note for the state:
  // if state == null  return IllegalArgumentException.
  // case 1: state = null and state = READY
  // - if task assigned to a user check this (authenticated) user can get the task.
  // - check that an other users of the candidates list can't get the task.
  // - if the task is not assigned (role mapper without performer assign)
  // - check that an other user in the candidate list can get the task.
  // - if 2 instances are created check that 2 tasks are available.
  // (check the tasks has READY state)
  // case 2: state = SUSPENDED
  // check task has been assigned to the user performing suspendTask
  // case 3: state = EXECUTING
  // check task has been assigned to the user performing startTask
  // case 4: state = FINISHED
  public void testGetTaskList() throws BonitaException, LoginException {

    String processId = "testGetTask";
    final String loginAdmin = "admin";
    final String passwordAdmin = "bpm";
    final String loginJohn = "john";
    final String passwordJohn = "bpm";
    URL xpdlUrl = this.getClass().getResource("testGetTask.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl, AdminsRoleMapper.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    TaskUUID taskUUID = null;
    ProcessInstanceUUID instanceUUID = null;
    ProcessInstanceUUID badInstanceUUID = IdFactory.getNewInstanceUUID();

    //**** case 1 test ***
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      //getToDoList with taskState = null
      try {
        getQueryRuntimeAPI().getTaskList(null);
        fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (IllegalArgumentException e) {
        //OK
      }
      try {
        getQueryRuntimeAPI().getTaskList(instanceUUID, null);
        fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (IllegalArgumentException e) {
        //OK
      }
      try {
        getQueryRuntimeAPI().getTaskList(instanceUUID, loginAdmin, null);
        fail("IllegalArgumentException has been raised by the QueryRuntimeAPIImpl !");
      } catch (IllegalArgumentException e) {
        //OK
      }

      //getToDoList to get task for activity = "a" with taskState = READY
      Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
      assertEquals(1, taskActivities.size());

      Collection<ActivityInstance< TaskInstance >> taskActivitiesWithInst =
        getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
      assertEquals(1, taskActivitiesWithInst.size());

      //InstanceNotFoundException for getTaskList with argt: instanceUUID, taskState
      try {
        getQueryRuntimeAPI().getTaskList(badInstanceUUID, TaskState.READY);
        fail("check InstanceNotFoundException has been raised");
      } catch (InstanceNotFoundException ie) {
        assertEquals(ie.getInstanceUUID(), badInstanceUUID);
      }

      //InstanceNotFoundException for getTaskList with argt: instanceUUID, userId, taskState
      try {
        getQueryRuntimeAPI().getTaskList(badInstanceUUID, loginAdmin, TaskState.READY);
        fail("check InstanceNotFoundException has been raised");
      } catch (InstanceNotFoundException ie) {
        assertEquals(ie.getInstanceUUID(), badInstanceUUID);
      }

      ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();
      assertNotNull(taskActivity.getBody().getUUID());
      taskUUID = taskActivity.getBody().getUUID();
      assertEquals("a", taskActivity.getActivityId());
      assertNotNull(taskActivity.getBody().getCreatedDate());
    } catch (ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }

    //check the user john can't get the task
    loginAs(loginJohn, passwordJohn);
    Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(0, taskActivities.size());

    // go on finish task a (create task b)
    // loginAdmin is the user assigned to the task : a
    loginAs(loginAdmin, passwordAdmin);

    //case 4: test for state = FINISHED
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertEquals(0, taskActivities.size());

    //start&finish task : a
    getRuntimeAPI().startTask(taskUUID, true);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.EXECUTING);
    assertEquals(1, taskActivities.size());
    getRuntimeAPI().finishTask(taskUUID, true);

    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertEquals(1, taskActivities.size());
    assertEquals("a", taskActivities.iterator().next().getActivityId());

    //check admin user can get the task b
    loginAs(loginAdmin, passwordAdmin);
    //Tasks for authenticated user
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());
    //Tasks for a choosen user (admin)
    Collection<ActivityInstance< TaskInstance >> taskActivitiesAdminUser =
      getQueryRuntimeAPI().getTaskList(loginAdmin,TaskState.READY);
    Collection<ActivityInstance< TaskInstance >> taskActivitiesAdminUserWithInstUUID =
      getQueryRuntimeAPI().getTaskList(instanceUUID, loginAdmin,TaskState.READY);

    assertEquals("Check loginAdmin can get its tasks",1, taskActivities.size());
    assertEquals("Check tasks can be got for admin user",1, taskActivitiesAdminUserWithInstUUID.size());

    ActivityInstance< TaskInstance > taskActivityAdminUser = taskActivitiesAdminUser.iterator().next();
    ActivityInstance< TaskInstance > taskActivity = taskActivities.iterator().next();

    assertEquals(taskActivityAdminUser.getActivityId() , taskActivity.getActivityId());

    //other checks on got taskInstance
    assertNotNull(taskActivity.getActivityId());
    taskUUID = taskActivity.getBody().getUUID();

    //check task b is not assigned and candidates list contains : admin, john
    assertEquals("SYSTEM", taskActivity.getBody().getUpdatedBy());
    assertFalse(taskActivity.getBody().isTaskAssigned());
    try {
      fail(taskActivity.getBody().getTaskUser());
    } catch (IllegalStateException ie) {
      //OK
    }


    assertTrue(taskActivity.getBody().getTaskCandidates().containsAll(Arrays.asList(loginAdmin, loginJohn)));
    assertEquals("b", taskActivity.getActivityId());

    //check john user can get also the task b
    loginAs(loginJohn, passwordJohn);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());

    //Check that if an other instance of the same process is created 2 tasks are available for admin user.
    loginAs(loginAdmin, passwordAdmin);
    instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
    //getToDoList to get task for activity = "a"
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(2, taskActivities.size());

    //delete the 2 nd created instance
    getRuntimeAPI().deleteProcessInstance(instanceUUID);

    //*** case 2 test ****
    //**suspend the task b withou assignment !  (only candidates are set)
    loginAs(loginAdmin, passwordAdmin);
    getRuntimeAPI().suspendTask(taskUUID, false);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.SUSPENDED);
    assertEquals(1, taskActivities.size());

    //check john user can't get the suspended task (by admin user)
    loginAs(loginJohn, passwordJohn);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.SUSPENDED);
    assertEquals(1, taskActivities.size());

    //check admin user can get the suspended task (by admin user)
    loginAs(loginAdmin, passwordAdmin);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.SUSPENDED);
    assertEquals(1, taskActivities.size());

    //**resumeTask and check that both admin and john can get the task b
    //meaning that the assigned user has been set as it was before the suspend (ie null)
    loginAs(loginJohn, passwordJohn);
    getRuntimeAPI().resumeTask(taskUUID, false);

    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());

    loginAs(loginAdmin, passwordAdmin);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertEquals(1, taskActivities.size());


    //*** case 3 test ***
    //check that only the user having done the startTask can get the task.
    loginAs(loginAdmin, passwordAdmin);
    getRuntimeAPI().startTask(taskUUID, true);
    loginAs(loginAdmin, passwordAdmin);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.EXECUTING);
    assertEquals(1, taskActivities.size());
    loginAs(loginJohn, passwordJohn);
    taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.EXECUTING);
    assertEquals(0, taskActivities.size());

    //go on until undeploying
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDoneList() throws BonitaException {
    String processId = "doneList";
    URL xpdlUrl = this.getClass().getResource("doneList.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        AdminsRoleMapper.class)).get(processId);
    ProcessDefinitionUUID processUUID = process.getUUID();
    
    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Collection<ActivityInstance<TaskInstance>> donelist = null;
    
    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    
    String otherUser = "anotherUser";
    
    //all tasks are assign to instanceInitiator by default
    donelist = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(0, donelist.size());
    
    todolist = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertNotNull(todolist);
    assertEquals(1, todolist.size());
    
    //execute the first task and check it appears in the done list
    TaskUUID task1UUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(task1UUID, true);
    getRuntimeAPI().finishTask(task1UUID, true);
    
    donelist = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(1, donelist.size());
    assertEquals(task1UUID, donelist.iterator().next().getBody().getUUID()); 
    
    //assign the second task to another user
    todolist = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertNotNull(todolist);
    assertEquals(1, todolist.size());
    TaskUUID task2UUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().assignTask(task2UUID, otherUser);
    
    //check the other user has task2 in its todolist
    todolist = getQueryRuntimeAPI().getTaskList(otherUser, TaskState.READY);
    assertNotNull(todolist);
    assertEquals(1, todolist.size());
    assertEquals(task2UUID, todolist.iterator().next().getBody().getUUID());
    
    //check the task is not anymore in todoList of the logged user
    todolist = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertNotNull(todolist);
    assertEquals(0, todolist.size());
    
    //execute the second task as 'another user"
    getRuntimeAPI().startTask(task2UUID, false);
    getRuntimeAPI().finishTask(task2UUID, false);
    
    //check the second task is  in the donelist of the other user
    donelist = getQueryRuntimeAPI().getTaskList(otherUser, TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(1, donelist.size());
    assertEquals(task2UUID, donelist.iterator().next().getBody().getUUID()); 
    
    //check the second task is not in the donelist of the current logged user. Only t1 is in donelist
    donelist = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(1, donelist.size());
    assertEquals(task1UUID, donelist.iterator().next().getBody().getUUID()); 
    
    //execute the third task and check it appears in the done list of the logged user
    todolist = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertNotNull(todolist);
    assertEquals(1, todolist.size());
    
    TaskUUID task3UUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(task3UUID, true);
    getRuntimeAPI().finishTask(task3UUID, true);
    
    donelist = getQueryRuntimeAPI().getTaskList(TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(2, donelist.size());
    
    //check doneList of the logged user contains task1 and task3
    Set<TaskUUID> expected = buildSet(task1UUID, task3UUID);
    Set<TaskUUID> donelistUUIDs = new HashSet<TaskUUID>();
    Iterator<ActivityInstance<TaskInstance>> it = donelist.iterator();
    while (it.hasNext()) {
      donelistUUIDs.add(it.next().getBody().getUUID());
    }
    assertEquals(expected, donelistUUIDs);
    
    //check doneList of the other user contains task2
    donelist = getQueryRuntimeAPI().getTaskList(otherUser, TaskState.FINISHED);
    assertNotNull(donelist);
    assertEquals(1, donelist.size());
    assertEquals(task2UUID, donelist.iterator().next().getBody().getUUID()); 
    
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  private Set<TaskUUID> buildSet(TaskUUID... taskUUIDs) {
    Set<TaskUUID> s = new HashSet<TaskUUID>();
    for (TaskUUID taskUUID : taskUUIDs) {
      s.add(taskUUID);
    }
    return s;
  }
  public void testGetTask() throws BonitaException {
    String processId = "usertest2";
    URL xpdlUrl = this.getClass().getResource("usertest2_1.0.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();
    TaskUUID newTaskUUID = IdFactory.getNewTaskUUID();
    ProcessInstanceUUID instanceUUID = null;
    try {
      instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
      assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());
      Collection<ActivityInstance< TaskInstance >> taskActivities = getQueryRuntimeAPI().getTaskList(TaskState.READY);
      assertEquals(1, taskActivities.size());
      TaskUUID taskUUID = taskActivities.iterator().next().getBody().getUUID();

      //getTask for task within activity = a
      ActivityInstance< TaskInstance > taskActivity = getQueryRuntimeAPI().getTask(taskUUID);
      assertNotNull(taskActivity.getActivityId());
      assertEquals("a", taskActivity.getActivityId());
      assertNotNull(taskActivity.getBody().getCreatedDate());
    } catch (ProcessNotFoundException be) {
      throw new RuntimeException("Missing resource!", be);
    }
    try {
      getQueryRuntimeAPI().getTask(newTaskUUID);
      fail("This task does not exist");
    } catch (TaskNotFoundException e) {
      assertEquals(newTaskUUID, e.getTaskUUID());
    }

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testDeleteAllProcessInstances() throws BonitaException {

    try {
      getRuntimeAPI().deleteAllProcessInstances(null);
    } catch (IllegalArgumentException e) {
      // Test pass
    }
  
    String processId = "deleteInstance";
    URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Set<ProcessInstance> instances = null;
    TaskUUID taskUUID = null;

    //check that both running and finished instances are deleted
    ProcessInstanceUUID instance1 = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance1, TaskState.READY);
    assertNotNull(todolist);
    assertFalse(todolist.isEmpty());
    assertTrue(todolist.size() == 1);
    
    ProcessInstanceUUID instance2 = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instance2).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance2, TaskState.READY);
    assertNotNull(todolist);
    assertFalse(todolist.isEmpty());
    assertTrue(todolist.size() == 1);
    taskUUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    assertEquals(InstanceState.COMPLETED, getQueryRuntimeAPI().getProcessInstance(instance2).getInstanceState());
    
    instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
    assertNotNull(instances);
    assertTrue(instances.size() == 2);
    
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    
    instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
    assertNotNull(instances);
    assertTrue(instances.isEmpty());
    
    //check that if we try to delete instances of an unknown process, processNotFoundException is thrown
    ProcessDefinitionUUID wrongProcessUUID = IdFactory.getNewProcessUUID();
    try {
      getRuntimeAPI().deleteAllProcessInstances(wrongProcessUUID);
      fail("An exception must be thrown when trying to delete instances of an unexisting process");
    } catch (ProcessNotFoundException pnfe) {
      // ok
    }
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    
    // Check UndeletableInstanceException is thrown when trying to delete process which have parentInstance
    process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get("subflow");
    packageUUID = process.getPackageDefinitionUUID();
    processUUID = process.getUUID();
    
    instance1 = getRuntimeAPI().instantiateProcess(processUUID);
    ProcessInstance processInstance = getQueryRuntimeAPI().getProcessInstance(instance1);
    Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    assertNotNull(instancesUUID);
    assertTrue(instancesUUID.size() == 1);
    ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next(); 
    instance2 = getRuntimeAPI().instantiateProcess(processUUID);
        
    try {
      getRuntimeAPI().deleteAllProcessInstances(
        getQueryRuntimeAPI().getProcessInstance(childInstanceUUID).getProcessDefinitionUUID());
      fail("Cannot delete a process instance having a parent instance");
    } catch (UndeletableInstanceException e) {
      // Ok
      instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
      assertNotNull(instances);
      assertTrue(instances.size() == 2);
    }
    
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  public void testDeleteProcessInstance() throws BonitaException {
    // many cases to test :
    // 0) check IllegalArgumentException with null parameter
    // 1) delete a running instance : check instance not available in querier
    // 2) delete an archived instance : check available in querier, delete, check not available
    // 3) check instanceNotFound thrown if instance can't be found in repository neither in journal, neither in history
    // 4) check UndeletableInstanceException

    // case 0
    try {
      getRuntimeAPI().deleteProcessInstance(null);
      fail("IllegalArgumentException must be thrown");
    } catch (IllegalArgumentException e) {
      // ok
    }

    String processId = "deleteInstance";
    URL xpdlUrl = this.getClass().getResource("deleteInstance.xpdl");
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instance1 = null;
    Collection<ActivityInstance<TaskInstance>> todolist = null;
    Set<ProcessInstance> instances = null;
    TaskUUID taskUUID = null;
    
    // case 1
    instance1 = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance1, TaskState.READY);
    assertNotNull(todolist);
    assertFalse(todolist.isEmpty());
    assertTrue(todolist.size() == 1);
    getRuntimeAPI().deleteProcessInstance(instance1);
    try {
      getQueryRuntimeAPI().getProcessInstance(instance1);
      fail("Instance is just deleted! We must a an exception here!");
    } catch (InstanceNotFoundException infe) {
      // ok
    }
    instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
    assertNotNull(instances);
    assertTrue(instances.isEmpty());
    
    // case 2
    instance1 = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(InstanceState.STARTED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    todolist = getQueryRuntimeAPI().getTaskList(instance1, TaskState.READY);
    assertNotNull(todolist);
    assertFalse(todolist.isEmpty());
    assertTrue(todolist.size() == 1);
    taskUUID = todolist.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    assertEquals(InstanceState.COMPLETED, getQueryRuntimeAPI().getProcessInstance(instance1).getInstanceState());
    
    getRuntimeAPI().deleteProcessInstance(instance1);
    try {
      getQueryRuntimeAPI().getProcessInstance(instance1);
      fail("Instance is just deleted! We must a an exception here!");
    } catch (InstanceNotFoundException infe) {
      // ok
    }
    instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
    assertNotNull(instances);
    assertTrue(instances.isEmpty());
    
    // case 3
    try {
      getRuntimeAPI().deleteProcessInstance(instance1);
      fail("Unable to delete an instance which can't be found in journal neither in repository");
    } catch (InstanceNotFoundException infe) {
      // ok
    }
    
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
    
    // case 4 : Check UndeletableInstanceException is thrown when trying to delete process which have parentInstance
    process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get("subflow");
    packageUUID = process.getPackageDefinitionUUID();
    processUUID = process.getUUID();
    
    instance1 = getRuntimeAPI().instantiateProcess(processUUID);
    ProcessInstance processInstance = getQueryRuntimeAPI().getProcessInstance(instance1);
    Set<ProcessInstanceUUID> instancesUUID = processInstance.getChildrenInstanceUUID();
    assertNotNull(instancesUUID);
    assertTrue(instancesUUID.size() == 1);
    ProcessInstanceUUID childInstanceUUID = instancesUUID.iterator().next(); 
    try {
      getRuntimeAPI().deleteProcessInstance(childInstanceUUID);
      fail("Cannot delete a process instance having a parent instance");
    } catch (UndeletableInstanceException e) {
      // Ok
      instances = getQueryRuntimeAPI().getProcessInstances(processUUID);
      assertNotNull(instances);
      assertTrue(instances.size() == 1);
      
      todolist = getQueryRuntimeAPI().getTaskList(TaskState.READY);
      assertNotNull(todolist);
      assertTrue(todolist.size() == 1);
      taskUUID = todolist.iterator().next().getBody().getUUID();
    
      getRuntimeAPI().startTask(taskUUID, true);
      getRuntimeAPI().finishTask(taskUUID, true);
    }
    
    getRuntimeAPI().deleteAllProcessInstances(processUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  private void loginAs(String loginName, String password) throws LoginException {
    loginContext.logout();
    try {
      loginContext = new LoginContext("Bonita", new SimpleCallbackHandler(loginName, password));
      loginContext.login();
    } catch (LoginException e) {
      throw new RuntimeException("Please, configure a JAAS test user with login: "
          + loginName + " and password: " + password, e);
    }
  }


}
