/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.integration.cycle;

import java.net.URL;
import java.util.Collection;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;

/**
 * Test. Deploys a xpdl process.
 *
 * @author Miguel Valdes
 */
public class CycleTest extends APITestCase {

  
  public void testBadIteration() {
    URL xpdlUrl = this.getClass().getResource("badIteration.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected DeploymentException");
    } catch (DeploymentException e) {
      // ok
      assertTrue(e.getMessage().contains("iteration points to an unknown node: cannot find a node with id 'a2'"));
    }
    
  }
  
  /*
   * Test a simple cycle.
   * Correct execution is: init a1 a2 a4 a1 a2 end
   */
  public void testSimpleCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("simpleCycle.xpdl");
    String processId = "simpleCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"init", "a4"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a cycle with a split/and and a Join/and
   * Correct execution is init -> a1 -> a3 -> a2 -> a4 -> a1 -> a3 -> a2 -> end
   *                               |--> a5 --|              |--> a5 --|
   */
  public void testSplitAndJoinCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("splitAndJoinCycle.xpdl");
    String processId = "splitAndJoinCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"init", "a4"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testSplitXorJoinCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("splitXorJoinCycle.xpdl");
    String processId = "splitXorJoinCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    int nb = 0;
    try {
      nb += getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size();
    } catch (ActivityNotFoundException e) {
      // TODO: handle exception
    }
    try {
      nb += getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size();
    } catch (ActivityNotFoundException e) {
      // TODO: handle exception
    }
    assertTrue(nb >= 2);
    assertTrue(nb <= 4);

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a simple proEd cycle.
   * Correct execution is: init a1 a2 a4 a1 a2 end
   */
  public void testSimpleCycleProEd() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("simpleCycleProEd.xpdl");
    String processId = "simpleCycleProEd";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"init", "a1"});

    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());

    TaskInstance task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkStopped(instanceUUID, new String[]{"init", "a2"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());

    task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a nested cycle.
   * Correct execution is: init a1 a3 a2 a4 a3 a2 a6 a5 a1 a3 a2 a6 end
   */
  public void testNestedCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("nestedCycle.xpdl");
    String processId = "nestedCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class, HookBeforeTerminateUpdateVariable2.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();


    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "a5", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(3, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    assertEquals(3, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a6").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }


  /*
   * Test an arbitrary cycle.
   * Correct execution is: init a1 a3 a2 a4 a1 a3 a2 a6 a5 a3 a2 a6 end
   */
  public void testArbitraryCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("arbitraryCycle.xpdl");
    String processId = "arbitraryCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class, HookBeforeTerminateUpdateVariable2.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "a5", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(3, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    assertEquals(3, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a6").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a simple proEd cycle with an extra entry point.
   * Correct execution is: init a3  a1 a2 a4 a1 a2 end
   */
  public void testEntryAndJoinProEd() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("entryAndCycle.xpdl");
    String processId = "entryAndCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"init", "a1", "a3"});

    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());

    TaskInstance task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkStopped(instanceUUID, new String[]{"init", "a2"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());

    task = tasks.iterator().next().getBody();
    getRuntimeAPI().startTask(task.getUUID(), true);
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
    getRuntimeAPI().finishTask(task.getUUID(), true);

    checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /*
   * Test a simple proEd cycle with an extra entry point.
   * Correct execution is: init a0/a3  a2 a1 a4 a2 a1 a4 a2 end
   */
  public void testEntryXorJoinProEd() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("entryXorCycle.xpdl");
    String processId = "entryXorCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkStopped(instanceUUID, new String[]{"init", "a1", "a2", "a3"});

    // Tasks from a4 and from a0
    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(2, tasks.size());
    TaskInstance task = null;
    for (ActivityInstance<TaskInstance> act : tasks) {
      if (act.getActivityId().equals("a4")) {
        task = act.getBody();
        getRuntimeAPI().startTask(task.getUUID(), true);
        getRuntimeAPI().finishTask(task.getUUID(), true);
      }
    }

    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(2, tasks.size());

    for (ActivityInstance<TaskInstance> act : tasks) {
      if (act.getActivityId().equals("a4")) {
        task = act.getBody();
        getRuntimeAPI().startTask(task.getUUID(), true);
        getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "cont", "0");
        getRuntimeAPI().finishTask(task.getUUID(), true);
      }
    }

    checkExecutedOnce(instanceUUID, new String[]{"init", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a4").size());
    assertEquals(3, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  /**
   * Checks that execution of a task in a previous iteration
   * (left there because of a Join XOR)
   * doesn't propagate to the new iteration.
   * @throws BonitaException
   */
  public void testJoinXorTaskCycle() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("joinXorTaskCycle.xpdl");
    String processId = "joinXorTaskCycle";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateUpdateVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(2, tasks.size());

    String iterationId = null;
    for (ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a3")) {
        getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
        iterationId = activity.getIterationId();
      }
    }
    assertNotNull(iterationId);

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(3, tasks.size());

    for (ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a5") && activity.getIterationId().equals(iterationId)) {
        getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
      }
    }

    checkStopped(instanceUUID, new String[]{"init", "a2"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(2, tasks.size());

    for (ActivityInstance<TaskInstance> activity : tasks) {
      if (activity.getActivityId().equals("a5")) {
        getRuntimeAPI().startTask(activity.getBody().getUUID(), true);
        getRuntimeAPI().finishTask(activity.getBody().getUUID(), true);
      }
    }

    checkExecutedOnce(instanceUUID, new String[]{"init", "a4", "BonitaEnd"});
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a1").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a2").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a3").size());
    assertEquals(2, getQueryRuntimeAPI().getActivityInstances(instanceUUID, "a5").size());

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  
  /*
   * Test a nested cycle with all nodes in the super cycle.
   * Correct execution is: init step0 step1 step2 step1 step2 step3 step0 step1 step2 step3 end
   */
  public void testNestedCycleNoExtraNodes() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("testNestedCycleNoExtraNodes.xpdl");
    String processId = "testNestedCycleNoExtraNodes";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();


    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    Collection<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    ActivityInstance<TaskInstance> task = tasks.iterator().next(); 
    assertEquals("init", task.getActivityId());
    TaskUUID taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step0", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "loop1", "yes");
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step3", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step0", task.getActivityId());
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "loop2", "yes");
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step1", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step2", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    tasks = getQueryRuntimeAPI().getTaskList(instanceUUID, TaskState.READY);
    assertEquals(1, tasks.size());
    task = tasks.iterator().next(); 
    assertEquals("step3", task.getActivityId());
    taskUUID = task.getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    getRuntimeAPI().finishTask(taskUUID, true);
    
    assertNotNull(getQueryRuntimeAPI().getProcessInstance(instanceUUID).getEndedDate());
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
}
