/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.integration.hook;

import java.net.URL;
import java.util.Set;

import junit.framework.Assert;
import junit.framework.AssertionFailedError;
import junit.framework.ComparisonFailure;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.UnAuthorizedUserException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.novabpm.util.Misc;

/**
 * Test. Deploys a xpdl process.
 *
 * @author Miguel Valdes
 */
public class ActivityHookTest extends APITestCase {

  @Override
  protected void setUp() throws Exception {
    super.setUp();
    getManagementAPI().deployClasses(getClasses(Assert.class, ComparisonFailure.class, AssertionFailedError.class));
  }
  
  @Override
  protected void tearDown() throws Exception {
    getManagementAPI().removeClass(Assert.class.getName());
    getManagementAPI().removeClass(ComparisonFailure.class.getName());
    getManagementAPI().removeClass(AssertionFailedError.class.getName());
    super.tearDown();
  }
  
  public void testActivityHookProEdV3() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("activityHook-v3.xpdl");
    activityHookTest(xpdlUrl);
  }

  public void testActivityHookProEdV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("activityHook-v4.xpdl");
    activityHookTest(xpdlUrl);
  }

  private void activityHookTest(URL xpdlUrl) throws DeploymentException,
  ProcessNotFoundException,
  InstanceNotFoundException,
  ActivityNotFoundException,
  UndeletablePackageException, 
  PackageNotFoundException, UndeletableInstanceException {
    String processId = "activityHook";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        HookBeforeTerminateCheckActVariable.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    checkExecutedOnce(instanceUUID, new String[] {"a", "b", "initial"});
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testOnReadyV3() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onReadyHook-v3.xpdl");
    onReadyTest(xpdlUrl);
  }

  public void testOnReadyV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onReadyHook-v4.xpdl");
    onReadyTest(xpdlUrl);
  }

  private void onReadyTest(final URL xpdlUrl) throws DeploymentException, ProcessNotFoundException,
  InstanceNotFoundException, ActivityNotFoundException,
  TaskNotFoundException, IllegalTaskStateException, 
  UndeletableInstanceException, PackageNotFoundException, 
  UndeletablePackageException {

    final String processId = "onReadyHook";
    final ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        AbstractHook.class, OnReadyHook.class)).get(processId);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = process.getUUID();

    final ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    final Set<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTasks(instanceUUID);
    assertTrue(tasks.size() == 1);
    Misc.deepToString(tasks);
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    getRuntimeAPI().startTask(task.getBody().getUUID(), true);
    getRuntimeAPI().finishTask(task.getBody().getUUID(), true);
    checkExecutedOnce(instanceUUID, new String[]{"firstActivity", "secondActivity"});
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testOnSuspendV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onSuspendHook-v4.xpdl");
    onSuspendTest(xpdlUrl);
  }

  private void onSuspendTest(URL xpdlUrl) throws DeploymentException, ProcessNotFoundException,
   InstanceNotFoundException, ActivityNotFoundException,
   TaskNotFoundException, IllegalTaskStateException, 
   UnAuthorizedUserException, UndeletableInstanceException, 
   PackageNotFoundException, UndeletablePackageException {

    String processId = "onSuspendHook";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        OnSuspendHook.class, AbstractHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    final Set<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTasks(instanceUUID);
    assertTrue(tasks.size() == 1);
    Misc.deepToString(tasks);
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    getRuntimeAPI().startTask(task.getBody().getUUID(), true);
    getRuntimeAPI().suspendTask(task.getBody().getUUID(), true);
    getRuntimeAPI().resumeTask(task.getBody().getUUID(), false);
    getRuntimeAPI().finishTask(task.getBody().getUUID(), true);
    checkExecutedOnce(instanceUUID, new String[]{"firstActivity", "secondActivity"});
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testOnResumeV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onResumeHook-v4.xpdl");
    onResumeTest(xpdlUrl);
  }

  private void onResumeTest(URL xpdlUrl) throws DeploymentException, ProcessNotFoundException,
   InstanceNotFoundException, ActivityNotFoundException, UndeletableInstanceException,
   TaskNotFoundException, IllegalTaskStateException, 
   UnAuthorizedUserException, PackageNotFoundException, UndeletablePackageException {

    String processId = "onResumeHook";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        OnResumeHook.class, AbstractHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    final Set<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTasks(instanceUUID);
    assertTrue(tasks.size() == 1);
    Misc.deepToString(tasks);
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    getRuntimeAPI().startTask(task.getBody().getUUID(), true);
    getRuntimeAPI().suspendTask(task.getBody().getUUID(), true);
    getRuntimeAPI().resumeTask(task.getBody().getUUID(), false);
    getRuntimeAPI().finishTask(task.getBody().getUUID(), true);
    checkExecutedOnce(instanceUUID, new String[]{"firstActivity", "secondActivity"});
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testOnStartV3() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onStartHook-v3.xpdl");
    onStartTest(xpdlUrl, "onStartHook");
  }

  public void testOnStartV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onStartHook-v4.xpdl");
    onStartTest(xpdlUrl, "onStartHook");
  }

  private void onStartTest(URL xpdlUrl, String processId) throws DeploymentException, ProcessNotFoundException,
  InstanceNotFoundException, ActivityNotFoundException,
  TaskNotFoundException, IllegalTaskStateException, 
  PackageNotFoundException, UndeletablePackageException, UndeletableInstanceException {

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        OnStartHook.class, AbstractHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    final Set<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTasks(instanceUUID);
    assertTrue(tasks.size() == 1);
    Misc.deepToString(tasks);
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    getRuntimeAPI().startTask(task.getBody().getUUID(), true);
    getRuntimeAPI().finishTask(task.getBody().getUUID(), true);
    checkExecutedOnce(instanceUUID, new String[]{"firstActivity", "secondActivity"});
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  public void testOnCancelV3() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onCancelHook-v3.xpdl");
    onCancelHookTest(xpdlUrl, "onCancel");
  }

  public void testOnCancelV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onCancelHook-v4.xpdl");
    onCancelHookTest(xpdlUrl, HookDefinition.Event.taskOnCancel.toString());
  }

  private void onCancelHookTest(URL xpdlUrl, String cancelName) {
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl, (Class[])null));
      fail(cancelName + " hook is not yet supported and no error is thrown at parsing.");
    } catch (DeploymentException e) {
      assertTrue("Bad error message: " + Misc.getStackTraceFrom(e),
          e.getMessage().contains("Unsupported HookEventName: " + cancelName));
    }
  }

  public void testOnFinishV3() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onFinishHook-v3.xpdl");
    onFinishHookTest(xpdlUrl);
  }

  public void testOnFinishV4() throws BonitaException {
    URL xpdlUrl = this.getClass().getResource("onFinishHook-v4.xpdl");
    onFinishHookTest(xpdlUrl);
  }

  private void onFinishHookTest(URL xpdlUrl) throws BonitaException {
    String processId = "onFinishHook";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl,
        OnFinishHook.class, AbstractHook.class)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    final Set<ActivityInstance<TaskInstance>> tasks = getQueryRuntimeAPI().getTasks(instanceUUID);
    assertTrue(tasks.size() == 1);
    Misc.deepToString(tasks);
    final ActivityInstance<TaskInstance> task = tasks.iterator().next();
    getRuntimeAPI().startTask(task.getBody().getUUID(), true);
    getRuntimeAPI().finishTask(task.getBody().getUUID(), true);
    checkExecutedOnce(instanceUUID, new String[]{"firstActivity", "secondActivity"});
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
}
