/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.services.record;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import junit.framework.TestCase;

import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.DateUtil;


/**
 * @author Guillaume Porcher
 *
 */
public abstract class AbstractTaskInstTest extends TestCase {

  protected static long currentTaskUUID = 0;
  protected static Object lock = new Object();
  protected final String userId = "james";
  protected final Set<String> candidates = new HashSet<String>();
  protected final TaskState state = TaskState.READY;
  // Use dates without milliseconds
  // Some DB does not support milliseconds
  protected final Date createdDate = DateUtil.parseDate("2008/08/25/09/57/42");

  public AbstractTaskInstTest() {
    super();
    candidates.add("john");
    candidates.add("jack");
  }

  protected abstract TaskInstance createTaskRecord(TaskUUID taskUUID, String userId, Set<String> candidates,
      TaskState state, Date createdDate);
  protected abstract void remove(TaskInstance taskInstance);
  protected abstract TaskInstance setStartedBy(TaskInstance taskInstance, String startUser);
  protected abstract TaskInstance setEndedBy(TaskInstance taskInstance, String endUser);
  protected abstract TaskInstance setStartedDate(TaskInstance taskInstance, Date startedTime);
  protected abstract TaskInstance setEndedDate(TaskInstance taskInstance, Date endedTime);


  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#
   * TaskFullInstanceImpl(java.lang.String, java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.util.Set,
   * java.lang.String, java.util.Date, java.util.Date)}.
   */
  public void testTaskInst() {
    // check that default constructor with non null args works.
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate);

    assertNotNull(taskInstance);
    remove(taskInstance);

    // Throw exception if taskUUID is null
    try {
      createTaskRecord(null, userId, candidates, state, createdDate);
      fail("An exception should be thrown if taskUUID is null");
    } catch (IllegalArgumentException e) {
      assertTrue(e.getMessage(),e.getMessage().contains("taskUUID"));
    }

    // Throw exception if state is null
    try {
      createTaskRecord(IdFactory.getNewTaskUUID(), userId, candidates, null, createdDate);
      fail("An exception should be thrown if initial state is null");
    } catch (IllegalArgumentException e) {
      assertTrue(e.getMessage().contains("initialState"));
    }

    // Throw exception if createdDate is null
    try {
      createTaskRecord(IdFactory.getNewTaskUUID(), userId, candidates, state, null);
      fail("An exception should be thrown if createdDate is null");
    } catch (IllegalArgumentException e) {
      assertTrue(e.getMessage().contains("createdDate"));
    }

    // Don't fail if other arguments are null.
    taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        null,
        candidates,
        state,
        createdDate);

    assertNotNull(taskInstance);
    remove(taskInstance);
  }


  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getCreatedDate()}.
   */
  public void testGetCreatedTime() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertEquals(createdDate, taskInstance.getCreatedDate());

    remove(taskInstance);
  }

  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getDueDate()}.
   */
  public void testGetDueDate() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertNull(taskInstance.getDueDate());
    remove(taskInstance);
  }

  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getEndedDate()}.
   */
  public void testGetEndedTime() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertNull(taskInstance.getEndedDate());

    Date startDate = new Date();
    Date endDate = new Date();

    taskInstance = setStartedDate(taskInstance, startDate);
    taskInstance = setEndedDate(taskInstance, endDate);

    assertEquals(endDate, taskInstance.getEndedDate());

    remove(taskInstance);
  }

  public void testGetEndedBy() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertNull(taskInstance.getEndedBy());

    String starteddBy = "Tom";
    String endedBy = "Joe";

    taskInstance = setStartedBy(taskInstance, starteddBy);
    taskInstance = setEndedBy(taskInstance, endedBy);

    assertEquals(endedBy, taskInstance.getEndedBy());

    remove(taskInstance);
  }


  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getStartedDate()}.
   */
  public void testGetStartedDate() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertNull(taskInstance.getStartedDate());

    Date startDate = new Date();

    taskInstance = setStartedDate(taskInstance, startDate);

    assertEquals(startDate, taskInstance.getStartedDate());

    remove(taskInstance);
  }

  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getStartedBy()}.
   */
  public void testGetStartedBy() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertNull(taskInstance.getStartedBy());

    String startUser = "john";

    taskInstance = setStartedBy(taskInstance, startUser);

    assertEquals(startUser, taskInstance.getStartedBy());

    remove(taskInstance);
  }

  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getPackageDefinitionUUID()}.
   */
  public void testGetId() {
    TaskUUID taskUUID = IdFactory.getNewTaskUUID();
    TaskInstance taskInstance = createTaskRecord(
        taskUUID,
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    assertEquals(taskUUID, taskInstance.getUUID());

    remove(taskInstance);
  }

  /**
   * Test method for {@link org.ow2.bonita.services.record.recording1.TaskFullInstanceImpl#getCandidates()}.
   */
  /*
   * getCandidates can be lazy loaded, taskRecord.getCandidates can throw an exception in db mode.
   * If no exception is raised, then the result must be the same as in memory mode.
   */
  public void testGetCandidates() {
    TaskInstance taskInstance = createTaskRecord(
        IdFactory.getNewTaskUUID(),
        userId,
        candidates,
        state,
        createdDate
    );

    assertNotNull(taskInstance);
    try {
      assertEquals(candidates, taskInstance.getTaskCandidates());
    } catch (RuntimeException e) {
      // Ok
    }

    remove(taskInstance);
  }

}
