/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.services.record;

import java.util.Date;
import java.util.Set;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.cmd.Command;
import org.jbpm.pvm.internal.cmd.CommandService;
import org.ow2.bonita.APITestCase;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.WorkflowElementFactory;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.IdFactory;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.services.Archivable;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Pierre Vigneras
 * @date Feb 27, 2008
 */
public class HistoryTest extends APITestCase {

  private static final int INSERTION_NB_MIN = 5;
  private static final int INSERTION_NB_MAX = 10;

  protected CommandService commandService;

  public void setUp() throws Exception {
    super.setUp();
    commandService = GlobalEnvironmentFactory.getEnvironmentFactory().get(CommandService.class);
  }

  public void tearDown() throws Exception {
    commandService = null;
    super.tearDown();
  }

  public void testArchiveSeveralTimeSamePackage() throws BonitaException {
    final String packageId = IdFactory.getNewPackageId();
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
      final long date = new Date().getTime();
      final PackageFullDefinition pack = WorkflowElementFactory.getNewPackage(packageUUID,
          new Date(date),
          new Date(date + 100),
          packageId);
      archivePackage(pack);
    }

    Set<PackageFullDefinition> records = getPackages();
    assertTrue(records.toString(), records.size() == n);
    records = getPackages(packageId);
    assertTrue(records.toString(), records.size() == n);
    for (PackageFullDefinition pack : records) {
      getManagementAPI().deletePackage(pack.getUUID());
    }
  }

  public void testArchiveDifferentPackages() throws BonitaException {
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      final long date = new Date().getTime();
      final String packageId = IdFactory.getNewPackageId();
      final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
      final PackageFullDefinition pack = WorkflowElementFactory.getNewPackage(packageUUID,
          new Date(date),
          new Date(date + 100),
          packageId);
      archivePackage(pack);
    }
    final Set<PackageFullDefinition> records = getPackages();
    assertTrue(records.toString(), records.size() == n);
    for (PackageFullDefinition pack : records) {
      getManagementAPI().deletePackage(pack.getUUID());
    }
  }

  public void testLastArchivedPackage() throws BonitaException {
    final String packageId = IdFactory.getNewPackageId();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final long date = new Date().getTime();
    final PackageFullDefinition record = WorkflowElementFactory.getNewPackage(packageUUID,
        new Date(date + 1000),
        new Date(date + 2000),
        packageId);
    archivePackage(record);
    final PackageDefinition last = getLastDeployedPackage(packageId, PackageDefinition.PackageState.UNDEPLOYED);
    assertNotNull(last);
    assertEquals(last.getPackageDefinitionUUID(), record.getPackageDefinitionUUID());

    final PackageDefinition pd = getPackage(record.getPackageDefinitionUUID());
    assertNotNull(Misc.deepToString(record), pd);
    assertEquals(Misc.deepToString(pd), record.getPackageDefinitionUUID(), pd.getPackageDefinitionUUID());
    getManagementAPI().deletePackage(record.getUUID());
  }

  public void testGetNonExistentPackage() {
    final PackageDefinition record = getLastDeployedPackage(IdFactory.getNewPackageId(), 
        PackageDefinition.PackageState.UNDEPLOYED);
    assertNull(record);
    final Set<PackageFullDefinition> records = getPackages(IdFactory.getNewPackageId());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetPackageNull() {
    try {
      getLastDeployedPackage(null, null);
      fail("Should not accept a null packageId!");
    } catch (IllegalArgumentException iae) {
      // ok
    }

    try {
      getPackages((String) null);
      fail("Should not accept a null packageId!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
  }

  public void testArchiveNullPackage() {
    try {
      archivePackage((PackageFullDefinition) null);
      fail("Should not accept a null packageFullDef!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
  }

  public void testArchiveSeveralInstances() throws BonitaException {
    ProcessInstanceUUID instanceUUID;
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();

    // Archive multiple instances of the same process
    final int n = Misc.random(INSERTION_NB_MIN, INSERTION_NB_MAX);
    for (int i = 0; i < n; i++) {
      instanceUUID = IdFactory.getNewInstanceUUID();
      final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
    }
    final Set<ProcessFullInstance> records = getProcessInstances(processUUID);
    assertTrue(records.size() + " == " + n + "?", records.size() == n);
    for (ProcessFullInstance instance : records) {
      getRuntimeAPI().deleteProcessInstance(instance.getUUID());  
    }
    
  }

  public void testArchiveOneInstanceAndCheckSomeFields() throws BonitaException {
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();
    final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID, packageUUID, processUUID);
    archiveProcess(instance);
    final ProcessFullInstance newRecord = getProcessInstance(instanceUUID);
    assertNotNull(newRecord);
    assertEquals(newRecord.getUUID(), instance.getUUID());
    assertEquals(newRecord.getProcessInstanceUUID(), instance.getProcessInstanceUUID());
    assertEquals(newRecord.getProcessDefinitionUUID(), instance.getProcessDefinitionUUID());
    assertEquals(newRecord.getParentInstanceUUID(), instance.getParentInstanceUUID());
    assertEquals(newRecord.getInitialVariableValues(), instance.getInitialVariableValues());
    getRuntimeAPI().deleteProcessInstance(instance.getUUID());
  }

  @SuppressWarnings("null")
  public void testArchiveSameInstance() throws BonitaException {
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();

    ProcessFullInstance instance = null;
    try {
      instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
      instance = WorkflowElementFactory.getNewInstance(instanceUUID,
          packageUUID,
          processUUID);
      archiveProcess(instance);
      fail("It should not be possible to archive twice the same instance!");
    } catch (IllegalArgumentException e) {
      // ok
      getRuntimeAPI().deleteProcessInstance(instance.getUUID());
    }
  }

  public void testArchiveNullProcessInstance() {
    try {
      archiveProcess((ProcessFullInstance) null);
      fail("Should not accept a null packageFullDef!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
  }

  public void testGetNonExistentProcessInstance() {
    ProcessFullInstance record = getProcessInstance(IdFactory.getNewInstanceUUID());
    assertNull(record);

    Set<ProcessFullInstance> records = getProcessInstances(IdFactory.getNewProcessUUID());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetProcessInstanceNull() {
    try {
      getProcessInstance(null);
      fail("Should not accept a null processInstanceUUID!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
    try {
      getProcessInstances(null);
      fail("Should not accept a null processInstanceUUID!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
  }

  @SuppressWarnings("unchecked")
  public void testActivity() throws BonitaException {
    final String activityId1 = IdFactory.getNewActivityId();
    final String activityId2 = IdFactory.getNewActivityId();
    final ActivityInstanceUUID activity1UUID = IdFactory.getNewActivityUUID();
    final ActivityInstanceUUID activity2UUID = IdFactory.getNewActivityUUID();
    final ProcessDefinitionUUID processUUID = IdFactory.getNewProcessUUID();
    final ProcessInstanceUUID instanceUUID = IdFactory.getNewInstanceUUID();
    final PackageDefinitionUUID packageUUID = IdFactory.getNewPackageUUID();
    final String iterationId = IdFactory.getNewIterationId();
    final String activityInstanceId = IdFactory.getNewActivityInstanceId();

    final ActivityFullInstance<TaskInstance> activity1 = WorkflowElementFactory.getNewActivity(activity1UUID,
        activityId1,
        packageUUID,
        processUUID,
        instanceUUID,
        iterationId,
        activityInstanceId);
    final ActivityFullInstance<TaskInstance> activity2 = WorkflowElementFactory.getNewActivity(activity2UUID,
        activityId2,
        packageUUID,
        processUUID,
        instanceUUID,
        iterationId,
        activityInstanceId);

    final ProcessFullInstance instance = WorkflowElementFactory.getNewInstance(instanceUUID,
        packageUUID,
        processUUID,
        activity1,
        activity2);

    archiveProcess(instance);

    final ProcessFullInstance historyInstance = getProcessInstance(instanceUUID);
    assertNotNull(historyInstance);
    final Set<ActivityFullInstance<ActivityBody>> instanceActivities = historyInstance.getActivities();
    assertNotNull(instanceActivities);
    assertEquals(2, instanceActivities.size());

    for (ActivityFullInstance<ActivityBody> historyActivity : instanceActivities) {
      final String activityId = historyActivity.getActivityId();
      if (activityId.equals(activityId1)) {
        assertActivityEquals(activity1, historyActivity);
      } else if (activityId.equals(activityId2)) {
        assertActivityEquals(activity2, historyActivity);
      } else {
        fail("Wrong Id: " + activityId);
      }
    }

    assertActivityEquals(activity1, getActivityInstance(instanceUUID, activityId1, iterationId, activityInstanceId));
    assertActivityEquals(activity2, getActivityInstance(instanceUUID, activityId2, iterationId, activityInstanceId));
    
    getRuntimeAPI().deleteProcessInstance(instance.getUUID());
  }

  public void testGetNonExistentActivityInstance() {
    Set<ActivityFullInstance<ActivityBody>> records;
    ActivityFullInstance<ActivityBody> record;
    record = getActivityInstance(IdFactory.getNewActivityUUID());
    assertNull(record);
    records = getActivityInstances(IdFactory.getNewInstanceUUID());
    assertNotNull(records);
    assertTrue(records.toString(), records.isEmpty());
  }

  public void testGetActivityInstanceNull() {
    try {
      getActivityInstance(null);
      fail("Should not accept a null activityInstanceUUID!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
    try {
      getActivityInstances(null);
      fail("Should not accept a null activityInstanceUUID!");
    } catch (IllegalArgumentException iae) {
      // ok
    }
  }

  private void assertActivityEquals(ActivityFullInstance< ? extends ActivityBody> expected,
      ActivityFullInstance< ? extends ActivityBody> actual) {
    assertNotNull(actual);
    assertNotNull(expected);
    assertEquals(expected.getUUID(), actual.getUUID());
    assertEquals(expected.getActivityId(), actual.getActivityId());
    assertEquals(expected.getProcessDefinitionUUID(), actual.getProcessDefinitionUUID());
    assertEquals(expected.getProcessInstanceUUID(), actual.getProcessInstanceUUID());
    assertEquals(expected.getIterationId(), actual.getIterationId());
    assertTrue(expected.getBody() instanceof TaskInstance);
    assertEquals(expected.getBody().getUUID(), actual.getBody().getUUID());
  }

  private void archiveProcess(ProcessFullInstance processInstance) {
    commandService.execute(new ArchiveCommand(processInstance));
  }

  private void archivePackage(PackageFullDefinition packageDef) {
    commandService.execute(new ArchiveCommand(packageDef));
  }

  private Set<PackageFullDefinition> getPackages() {
    return commandService.execute(new GetPackagesCommand());
  }

  private Set<PackageFullDefinition> getPackages(String packageId) {
    return commandService.execute(new GetPackagesCommandWithId(packageId));
  }

  private PackageFullDefinition getLastDeployedPackage(String packageId, PackageDefinition.PackageState packageState) {
    return commandService.execute(new GetLastDeployedPackage(packageId, packageState));
  }

  private ActivityFullInstance<ActivityBody> getActivityInstance(ActivityInstanceUUID instanceUUID) {
    return commandService.execute(new GetActivityInstanceOnlyUUID(instanceUUID));
  }

  private ActivityFullInstance<ActivityBody> getActivityInstance(ProcessInstanceUUID instanceUUID,
      String activityId, String iterationId, String activityInstanceId) {
    return commandService.execute(new GetActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId));
  }

  private Set<ActivityFullInstance<ActivityBody>> getActivityInstances(ProcessInstanceUUID processInstanceUUID) {
    return commandService.execute(new GetActivityInstances(processInstanceUUID));
  }

  private ProcessFullInstance getProcessInstance(ProcessInstanceUUID processInstanceUUID) {
    return commandService.execute(new GetProcessInstance(processInstanceUUID));
  }

  private Set<ProcessFullInstance> getProcessInstances(ProcessDefinitionUUID processDefUUID) {
    return commandService.execute(new GetProcessInstances(processDefUUID));
  }

  private PackageFullDefinition getPackage(PackageDefinitionUUID packDefUUID) {
    return commandService.execute(new GetPackageWithUUID(packDefUUID));
  }
  private Querier getHistory() {
    return EnvTool.getHistoryQueriers();
  }
  /* List of commands */

  class GetPackageWithUUID implements Command<PackageFullDefinition> {
    private PackageDefinitionUUID packDefUUID;

    public GetPackageWithUUID(PackageDefinitionUUID packDefUUID) {
      this.packDefUUID = packDefUUID;
    }
    public PackageFullDefinition execute(Environment environment) throws Exception {
      return getHistory().getPackage(packDefUUID);
    }
  }

  class GetProcessInstances implements Command<Set<ProcessFullInstance>> {
    private ProcessDefinitionUUID processDefUUID;

    public GetProcessInstances(ProcessDefinitionUUID processDefUUID) {
      this.processDefUUID = processDefUUID;
    }
    public Set<ProcessFullInstance> execute(Environment environment) throws Exception {
      return getHistory().getProcessInstances(processDefUUID);
    }
  }

  class GetProcessInstance implements Command<ProcessFullInstance> {
    private ProcessInstanceUUID processInstanceUUID;

    public GetProcessInstance(ProcessInstanceUUID instanceUUID) {
      this.processInstanceUUID = instanceUUID;
    }
    public ProcessFullInstance execute(Environment environment) throws Exception {
      ProcessFullInstance processInstance = getHistory().getProcessInstance(processInstanceUUID);
      if (processInstance != null) {
        // Load process instance attributes to prevent LazyInitializationException
        for (Object o : processInstance.getInitialVariableValues().values()) {
          o.toString();
        }
        for (Object o : processInstance.getLastKnownVariableValues().values()) {
          o.toString();
        }
        for (VariableUpdate o : processInstance.getVariableUpdates()) {
          o.toString();
        }        
        for (ActivityFullInstance<ActivityBody> activity : processInstance.getActivities()) {
          activity.toString();
        }
      }
      return processInstance;
    }

  }

  class GetActivityInstances implements Command<Set<ActivityFullInstance<ActivityBody>>> {
    private ProcessInstanceUUID processInstanceUUID;

    public GetActivityInstances(ProcessInstanceUUID instanceUUID) {
      this.processInstanceUUID = instanceUUID;
    }
    public Set<ActivityFullInstance<ActivityBody>> execute(Environment environment) throws Exception {
      return getHistory().getActivityInstances(processInstanceUUID);
    }

  }

  class GetActivityInstance implements Command<ActivityFullInstance<ActivityBody>> {

    private ProcessInstanceUUID processInstanceUUID;
    private String activityId;
    private String iterationId;
    private String activityInstanceId;

    public GetActivityInstance(ProcessInstanceUUID instanceUUID,
        String activityId, String iterationId, String activityInstanceId) {
      this.processInstanceUUID = instanceUUID;
      this.activityId = activityId;
      this.iterationId = iterationId;
      this.activityInstanceId = activityInstanceId;
    }
    public ActivityFullInstance<ActivityBody> execute(Environment environment) throws Exception {
      return getHistory().getActivityInstance(processInstanceUUID, activityId, iterationId, activityInstanceId);
    }

  }

  class GetActivityInstanceOnlyUUID implements Command<ActivityFullInstance<ActivityBody>> {
    private ActivityInstanceUUID actInstanceUUID;

    public GetActivityInstanceOnlyUUID(ActivityInstanceUUID instanceUUID) {
      this.actInstanceUUID = instanceUUID;
    }
    public ActivityFullInstance<ActivityBody> execute(Environment environment) throws Exception {
      return getHistory().getActivityInstance(actInstanceUUID);
    }

  }

  class GetLastDeployedPackage implements Command<PackageFullDefinition> {

    private String packageId;
    private PackageDefinition.PackageState packageState;

    public GetLastDeployedPackage(String packageId, PackageDefinition.PackageState packageState) {
      this.packageId = packageId;
      this.packageState = packageState;
    }
    public PackageFullDefinition execute(Environment environment) throws Exception {
      PackageFullDefinition pack = getHistory().getLastDeployedPackage(packageId, packageState);
      return pack;
    }
  }

  class GetPackagesCommand implements Command<Set<PackageFullDefinition>> {

    public Set<PackageFullDefinition> execute(Environment environment) throws Exception {
      Set<PackageFullDefinition> packages = getHistory().getPackages();
      if (packages != null) {
        //Load packages to prevent LazyInitializationException
        packages.toString();
      }
      return packages;
    }
  }

  class GetPackagesCommandWithId implements Command<Set<PackageFullDefinition>> {

    private String packageId = null;

    public GetPackagesCommandWithId(String packageId) {
      this.packageId = packageId;
    }

    public Set<PackageFullDefinition> execute(Environment environment) throws Exception {
      Set<PackageFullDefinition> packages = getHistory().getPackages(packageId);
      if (packages != null) {
        //Load packages to prevent LazyInitializationException
        packages.toString();
      }
      return packages;
    }
  }

  class ArchiveCommand implements Command<Archivable> {

    private PackageFullDefinition packageDef = null;
    private ProcessFullInstance processInstance = null;

    public ArchiveCommand(PackageFullDefinition packageDef) {
      this.packageDef = packageDef;
    }
    public ArchiveCommand(ProcessFullInstance processInstance) {
      this.processInstance = processInstance;
    }

    public Archivable execute(Environment environment) throws Exception {
      Archiver archiver = EnvTool.getArchiver();
      if (packageDef != null) {
        archiver.archive(packageDef);
        return packageDef;
      }
      archiver.archive(processInstance);
      return processInstance;
    }
  }
}
