package org.ow2.bonita.util;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.type.Variable;
import org.jbpm.pvm.internal.type.variable.StringVariable;
import org.jbpm.pvm.internal.type.variable.UnpersistableVariable;
import org.jmock.Expectations;
import org.jmock.Mockery;
import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.var.Enumeration;
import org.ow2.bonita.runtime.var.EnumerationVariable;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Recorder;
import org.ow2.novabpm.util.ClassDataTool;
import org.ow2.novabpm.util.Command;

public class ToolTest extends APITestCase {

  private Mockery journalContext = null;
  private Mockery archiverContext = null;
  private Mockery archivableContext = null;

  @Override
  protected void setUp() throws Exception {
    // TODO Auto-generated method stub
    super.setUp();
    this.journalContext = new Mockery();
    this.archiverContext = new Mockery();
    this.archivableContext = new Mockery();
  }

  @Override
  protected void tearDown() throws Exception {
    this.journalContext = null;
    this.archiverContext = null;
    this.archivableContext = null;
    super.tearDown();
  }


  public void testCreateVariableMap() {
    assertNull(FullCopyTool.createVariableMap(new HashMap<String, Object>()));

    assertNull(FullCopyTool.createVariableMap(null));
  }

  public void testGetClasses() throws IOException {
    assertEquals(new HashSet<byte[]>(), ClassDataTool.getClasses());

    final Class< ? > nullClass = null;
    try {
      ClassDataTool.getClasses(nullClass);
      fail("Exception should be throw when calling getClasses with a null class");
    } catch (final IOException e) {
      // Test passed
    }
  }

  public void testGetClassData() throws IOException {
    try {
      ClassDataTool.getClassData((Class < ? >)null);
      fail("BonitaRuntimeException should be thrown when calling getClassData with null parameter");
    } catch (final IOException e) {
      // Test passed
    }

    try {
      ClassDataTool.getClassData(ToolTest.class);
    } catch (final BonitaRuntimeException e) {
      fail("BonitaRuntimeException must not be thrown");
    }
  }

  // Journal and Archiver should not be null
  public void testAtomicArchiveNullParameters() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);
    try {
      ArchiveTool.atomicArchive(null, journal, archiver);
      fail("atomic archive should'nt passed with null parameters");
    } catch (final IllegalArgumentException e) {
      // Test passed
    }
  }

  public void testAtomicArchiveInvalidArchivable() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final ArchivableTest archivableTest = new ArchivableTest();

    this.journalContext.checking(new Expectations() {
      { never(journal); }
    });
    this.archiverContext.checking(new Expectations() {
      { never(archiver); }
    });

    try {
      ArchiveTool.atomicArchive(archivableTest, journal, archiver);
      fail("atomic archive should'nt passed with an invalid archivable");
    } catch (final IllegalArgumentException e) {
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    }
  }

  public void testAtomicArchiveArchiveErrorProcessFullInstance() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final ProcessFullInstance processDef = this.archivableContext.mock(ProcessFullInstance.class);

    this.journalContext.checking(new Expectations() {
      {
        never(journal).remove(processDef);
      }
    });
    this.archiverContext.checking(new Expectations() {
      {
        one(archiver).archive(processDef);
        will(throwException(new RuntimeException("archive exception")));
      }
    });
    this.archivableContext.checking(new Expectations() {
      {
        ignoring(processDef);
      }
    });

    try {
      ArchiveTool.atomicArchive(processDef, journal, archiver);
      fail("atomicArchive should'nt passed when archive don't pass");
    } catch (final AtomicArchiveException e) {
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    }
  }

  public void testAtomicArchiveArchiveErrorPackageFullDefinition() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final PackageFullDefinition packageDef = this.archivableContext.mock(PackageFullDefinition.class);

    this.journalContext.checking(new Expectations() {
      {
        never(journal).remove(packageDef);
      }
    });
    this.archiverContext.checking(new Expectations() {
      {
        one(archiver).archive(packageDef);
        will(throwException(new RuntimeException("archive exception")));
      }
    });
    this.archivableContext.checking(new Expectations() {
      {
        ignoring(packageDef);
      }
    });

    try {
      ArchiveTool.atomicArchive(packageDef, journal, archiver);
      fail("atomicArchive should'nt passed when archive don't pass");
    } catch (final AtomicArchiveException e) {
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    }
  }

  public void testAtomicArchiveArchivePassProcessFullInstance() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final ProcessFullInstance processDef = this.archivableContext.mock(ProcessFullInstance.class);

    this.journalContext.checking(new Expectations() {
      {
        one(journal).remove(processDef);
      }
    });
    this.archiverContext.checking(new Expectations() {
      {
        one(archiver).archive(processDef);
      }
    });
    this.archivableContext.checking(new Expectations() {
      {
        ignoring(processDef);
      }
    });

    try {
      ArchiveTool.atomicArchive(processDef, journal, archiver);
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    } catch (final AtomicArchiveException e) {
      fail("atomicArchive should pass");
    }
  }

  public void testAtomicArchiveArchivePassPackageFullDefinition() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final PackageFullDefinition packageDef = this.archivableContext.mock(PackageFullDefinition.class);

    this.journalContext.checking(new Expectations() {
      {
        one(journal).remove(packageDef);
      }
    });
    this.archiverContext.checking(new Expectations() {
      {
        one(archiver).archive(packageDef);
      }
    });

    this.archivableContext.checking(new Expectations() {
      {
        ignoring(packageDef);
      }
    });

    try {
      ArchiveTool.atomicArchive(packageDef, journal, archiver);
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    } catch (final AtomicArchiveException e) {
      fail("atomicArchive should pass");
    }
  }

  public void testAtomicArchiveJournalError() {
    final Recorder journal = this.journalContext.mock(Recorder.class);
    final Archiver archiver = this.archiverContext.mock(Archiver.class);

    final ProcessFullInstance archivable = this.archivableContext.mock(ProcessFullInstance.class);

    this.journalContext.checking(new Expectations() {
        {
          one(journal).remove(archivable);will(throwException(new RuntimeException()));
        }
      });
    this.archiverContext.checking(new Expectations() {
      {
        one(archiver).archive(archivable);
      }
    });

    this.archivableContext.checking(new Expectations() {
      {
        ignoring(archivable);
      }
    });

    try {
      ArchiveTool.atomicArchive(archivable, journal, archiver);
      fail("atomicArchive shouldn't pass");
    } catch (final AtomicArchiveException e) {
      this.journalContext.assertIsSatisfied();
      this.archiverContext.assertIsSatisfied();
    }
  }

  private static class UnsupportedVariable { }

  public void testCreateVariableIncorrect() throws Exception {
    getCommandAPI().execute(new Command<Object>() {
      public Object execute(final Environment environment) throws Exception {
        final Object u = new UnsupportedVariable();
        final Variable v = FullCopyTool.createVariable("int", u);
        assertTrue(v instanceof UnpersistableVariable);
        return null;
      }
    });
  }

  public void testCreateVariableString() throws Exception {
    getCommandAPI().execute(new Command<Object>() {
      public Object execute(final Environment environment) throws Exception {
        final Variable v = FullCopyTool.createVariable("string", "string");
        assertTrue(v instanceof StringVariable);
        return null;
      }
    });
  }

  public void testCreateVariableEnum() throws Exception {
    getCommandAPI().execute(new Command<Object>() {
      public Object execute(final Environment environment) throws Exception {
        final HashSet<String> possibleValues = new HashSet<String>();
        possibleValues.add("type1");
        possibleValues.add("type2");
        final Enumeration e = new Enumeration(possibleValues, "type1");
        assertTrue(FullCopyTool.createVariable("enum", e) instanceof EnumerationVariable);
        return null;
      }
    });
  }
}
