/**
 * Copyright (C) 2007 Bull S. A. S. Bull, Rue Jean Jaures, B.P.68, 78340, Les
 * Clayes-sous-Bois This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation version 2.1 of the License. This
 * library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details. You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
package org.ow2.bonita.variable;

import java.net.URL;
import java.util.Collection;

import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.util.BonitaException;

/**
 * Testing activity variables (string and enumeration types)
 * by using StartMode=Manual (task behavior)
 * taking in account :
 *     propagation parameter,
 *     split, join for execution
 * Variables are defined either into WFProcess or Package
 */
public class ActivityVariableTest extends VariableTestCase {

  // check that activity variables (string and enumeration) defined
  // into  WFProcess are :
  //      - available within the execution pointing to node with the given activity Id
  //        according the extended attribute
  //      - propagated according the extended attribute

  public void testActivityVariablesIntoWFProcess() throws BonitaException {
    URL xpdlUrl = ActivityVariableTest.class.getResource("varActivityIntoWFProcess.xpdl");
    String processId = "varActivityIntoWFProcess";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);


    // Check variables within execution pointing to act1 node
    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act1");
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    
    assertTrue(getQueryRuntimeAPI().getActivityInstanceVariables(activityInst.getUUID()).isEmpty());
    checkStopped(instanceUUID, new String[]{});

    // start & terminate "act1" task
    executeTask(instanceUUID, "act1");
    checkStopped(instanceUUID, new String[]{"act1"});
    
    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act2");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);

    // Check variables within execution pointing to act2 node
    checkVariables(getQueryRuntimeAPI(), activityInst.getUUID(), 7);

    // start & terminate "act2" task
    executeTask(instanceUUID, "act2");
    // check variables within execution pointing to act3 node
    acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "act3");
    assertEquals(1, acts.size());
    activityInst = acts.iterator().next();
    assertNotNull(activityInst);

    checkVariables(getQueryRuntimeAPI(), activityInst.getUUID(), 0);
    checkStopped(instanceUUID, new String[]{"act1", "act2"});

    // start & terminate "act3" task
    executeTask(instanceUUID, "act3");
    checkExecutedOnce(instanceUUID, new String[]{"act1", "act2", "act3"});

    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().undeploy(packageUUID);
    getManagementAPI().deletePackage(packageUUID);
  }


  // check that activity variables (string and enumeration) defined
  // into  Package are :
  //      - available within the execution pointing to node with the given activity Id
  //        according the extended attribute
  //      - propagated according the extended attribute

  public void testActivityVariablesIntoPackage() throws BonitaException {
    URL xpdlUrl = ActivityVariableTest.class.getResource("varActivityIntoPackage.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("This test was expected to fail at deploy time as it is trying to declare an activity datafield at package level....");
    } catch (DeploymentException e) {
      String start = "error : A datafield with processDefinitionUUID = ";
      String end = " defined at package level is expected to be an 'activity datafield' "
        + "as it declares the corresponding extended attribute: it is forbiden. "
        + "'Activity datafields' can only be defined at process level.";
      assertTrue(e.getMessage().contains(start + "package_str_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_str_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_flt_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_flt_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_int_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_int_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_bool_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_bool_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_date_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_date_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_perf_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_perf_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_enum_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "package_enum_act_no_propagation" + end));

      start = "error : Looking for a datafield with processDefinitionUUID: ";
      end = " in enclosing process but unable to find it.";
      assertTrue(e.getMessage().contains(start + "str_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "str_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act_no_propagation" + end));

    }
  }

  // check that activity variables (string and enumeration) defined
  // into  PACKAGE are well managed (including propagation)
  //      -  when spliting the initial execution
  //      -  when joining is performed

  public void testActivityPackageVariableswithSplitJoin() throws BonitaException {
    URL xpdlUrl = ActivityVariableTest.class.getResource("varActivityIntoPackageWithSplitJoin.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected exception:  BonitaException has not been raised");
    } catch (DeploymentException e) {
      String start = "error : A datafield with processDefinitionUUID = ";
      String end = " defined at package level is expected to be an 'activity datafield' "
        + "as it declares the corresponding extended attribute: it is forbiden. "
        + "'Activity datafields' can only be defined at process level.";
      assertTrue(e.getMessage().contains(start + "str_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "str_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act2_no_propagation" + end));

      assertTrue(e.getMessage().contains(start + "str_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "str_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act3_no_propagation" + end));

      start = "error : Looking for a datafield with processDefinitionUUID: ";
      end = " in enclosing process but unable to find it.";
      assertTrue(e.getMessage().contains(start + "str_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "str_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act2_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act2_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act2_no_propagation" + end));

      assertTrue(e.getMessage().contains(start + "str_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "str_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "flt_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "int_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "bool_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "date_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "perf_act3_no_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act3_propagation" + end));
      assertTrue(e.getMessage().contains(start + "enum_act3_no_propagation" + end));
    }

  }

}
