/**
 * Copyright (C) 2007 Bull S. A. S. Bull, Rue Jean Jaures, B.P.68, 78340, Les
 * Clayes-sous-Bois This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation version 2.1 of the License. This
 * library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details. You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */
package org.ow2.bonita.variable;

import java.net.URL;
import java.util.Collection;
import java.util.Date;

import org.ow2.bonita.APITestCase;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.ExpressionEvaluationException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.runtime.var.Enumeration;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.DateUtil;

/**
 * Testing process variables (string and enumeration types)
 *
 */
public class ProcessVariableTest extends APITestCase {

  /*
   * check the creation of process variable defined within the WorkflowProcess:
   *    - string type with null/no/empty/valid initialValue
   *    - enumeration type with initialValue
   */
  public void testProcessVariablesIntoWFProcess() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varProcess.xpdl");
    String processId = "varProcess";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkExecutedOnce(instanceUUID, new String[] {});
    // Check variables have been set within the instance
    checkStringVariable(getQueryRuntimeAPI(), instanceUUID);
    checkFloatVariable(getQueryRuntimeAPI(), instanceUUID);
    checkIntegerVariable(getQueryRuntimeAPI(), instanceUUID);
    checkBooleanVariable(getQueryRuntimeAPI(), instanceUUID);
    checkDateVariable(getQueryRuntimeAPI(), instanceUUID);
    checkPerformerVariable(getQueryRuntimeAPI(), instanceUUID);
    checkEnumerationVariable(getQueryRuntimeAPI(), instanceUUID);

    getManagementAPI().undeploy(packageUUID);
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().deletePackage(packageUUID);
  }

  // Check that a process variable is not seen as an activity variable
  public void testProcessVariablesNotLocal() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varProcess.xpdl");
    String processId = "varProcess";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkExecutedOnce(instanceUUID, new String[] {});

    // Process variable is defined
    String str = (String) getQueryRuntimeAPI().getProcessInstanceVariable(instanceUUID, "str_init");
    assertNotNull(str);
    assertEquals("initial value", str);
    // ActivityVariable is undefined
    Collection<ActivityInstance<ActivityBody>> acts = getQueryRuntimeAPI().getActivityInstances(instanceUUID, "r1");
    assertEquals(1, acts.size());
    ActivityInstance<ActivityBody> activityInst = acts.iterator().next();
    assertNotNull(activityInst);
    ActivityInstanceUUID activityUUID = activityInst.getUUID();
    try {
      getQueryRuntimeAPI().getActivityInstanceVariable(activityUUID, "str_init");
      fail("expected exception: " + VariableNotFoundException.class.getName());
    } catch (VariableNotFoundException e) {
      assertEquals("str_init", e.getVariableId());
    }
    // getVariable returns the global variable
    str = (String) getQueryRuntimeAPI().getVariable(activityUUID, "str_init");
    assertNotNull(str);
    assertEquals("initial value", str);
    
    getManagementAPI().undeploy(packageUUID);
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  /*
   * check the creation of process variable defined within the package:
   *    - string type with null/no/empty/valid initialValue
   *    - enumeration type with initialValue
   */
  public void testProcessVariablesIntoPackage() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varPackage.xpdl");
    String processId = "varPackage";

    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);

    checkExecutedOnce(instanceUUID, new String[] {});

    // Check variables have been set within the instance
    checkStringVariable(getQueryRuntimeAPI(), instanceUUID);
    checkFloatVariable(getQueryRuntimeAPI(), instanceUUID);
    checkIntegerVariable(getQueryRuntimeAPI(), instanceUUID);
    checkBooleanVariable(getQueryRuntimeAPI(), instanceUUID);
    checkDateVariable(getQueryRuntimeAPI(), instanceUUID);
    checkPerformerVariable(getQueryRuntimeAPI(), instanceUUID);
    checkEnumerationVariable(getQueryRuntimeAPI(), instanceUUID);

    getManagementAPI().undeploy(packageUUID);
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().deletePackage(packageUUID);

  }

  /*
   *   test an exception is raised for enumeration type if initial value element is null
   *   (<InitialValue/>)
   */
  public void testProcessEnumNullInit() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varEnumNullInit.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected exception:  BonitaException has not been raised");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains("Invalid enumeration value"));
    }
  }

  /*
   *   test an exception is raised for enumeration type if there's no initial value element
   */
  public void testProcessEnumNoInit() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varEnumNoInit.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected exception:  BonitaException has not been raised");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains("Invalid enumeration value"));
    }
  }

  /*
   *   test an exception is raised for enumeration type if initial value element is empty
   *   (<InitialValue></InitialValue>)
   */
  public void testProcessEnumEmptyInit() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("varEnumEmptyInit.xpdl");
    try {
      getManagementAPI().deployBar(getBarContent(xpdlUrl));
      fail("Expected exception:  BonitaException has not been raised");
    } catch (DeploymentException e) {
      assertTrue(e.getMessage(), e.getMessage().contains("Invalid enumeration value"));
    }
  }

  public void testNoInitialValue() throws BonitaException {
    URL xpdlUrl = ProcessVariableTest.class.getResource("noInitialValue.xpdl");
    String processId = "noInitialValue";
    ProcessDefinition process = getManagementAPI().deployBar(getBarContent(xpdlUrl)).get(processId);
    PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    ProcessDefinitionUUID processUUID = process.getUUID();

    ProcessInstanceUUID instanceUUID = getRuntimeAPI().instantiateProcess(processUUID);
    Collection<ActivityInstance<TaskInstance>> todoList = getQueryRuntimeAPI().getTaskList(TaskState.READY);
    assertNotNull(todoList);
    assertFalse(todoList.isEmpty());
    assertTrue(todoList.size() == 1);
    TaskUUID taskUUID = todoList.iterator().next().getBody().getUUID();
    getRuntimeAPI().startTask(taskUUID, true);
    try {
      getRuntimeAPI().finishTask(taskUUID, true);
      fail("An expression is using a null var. Must throw an exception");
    } catch (ExpressionEvaluationException e) {
      //check the message contains the var name
      assertTrue(e.getMessage().contains("nullVar"));
    }
    getRuntimeAPI().setProcessInstanceVariable(instanceUUID, "nullVar", "ok");
    getRuntimeAPI().finishTask(taskUUID, true);
    assertEquals(InstanceState.COMPLETED, getQueryRuntimeAPI().getProcessInstance(instanceUUID).getInstanceState());

    getManagementAPI().undeploy(packageUUID);
    getRuntimeAPI().deleteProcessInstance(instanceUUID);
    getManagementAPI().deletePackage(packageUUID);
  }
  private void checkStringVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // str_no_init
    String str = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "str_no_init");
    assertNull(str);

    // str_null_init
    // can be null in oracle
    str = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "str_null_init");
    assertTrue(str == null || "".equals(str));

    // str_empty_init
    // can be null in oracle
    str = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "str_empty_init");
    assertTrue(str == null || "".equals(str));


    // str_init
    str = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "str_init");
    assertNotNull(str);
    assertEquals("initial value", str);
  }

  private void checkFloatVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // flt_no_init
    Double flt = (Double) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "flt_no_init");
    assertNull(flt);

    // flt_null_init
    flt = (Double) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "flt_null_init");
    assertNotNull(flt);
    assertEquals((double)0, flt);

    // flt_empty_init
    flt = (Double) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "flt_empty_init");
    assertNotNull(flt);
    assertEquals((double)0, flt);

    // flt_init
    flt = (Double) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "flt_init");
    assertNotNull(flt);
    assertEquals(42.42, flt);
  }

  private void checkIntegerVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // int_no_init
    Long i = (Long) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "int_no_init");
    assertNull(i);

    // int_null_init
    i = (Long) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "int_null_init");
    assertNotNull(i);
    assertEquals(0L, (long)i);

    // int_empty_init
    i = (Long) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "int_empty_init");
    assertNotNull(i);
    assertEquals(0L, (long)i);

    // int_init
    i = (Long) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "int_init");
    assertNotNull(i);
    assertEquals(42L, (long)i);
  }

  private void checkBooleanVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // bool_no_init
    Boolean b = (Boolean) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "bool_no_init");
    assertNull(b);

    // bool_null_init
    b = (Boolean) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "bool_null_init");
    assertNotNull(b);
    assertEquals(false, (boolean)b);

    // bool_empty_init
    b = (Boolean) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "bool_empty_init");
    assertNotNull(b);
    assertEquals(false, (boolean)b);

    // bool_init
    b = (Boolean) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "bool_init");
    assertNotNull(b);
    assertEquals(true, (boolean)b);
  }

  private void checkDateVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // date_no_init
    Date d = (Date) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "date_no_init");
    assertNull(d);

    // date_null_init
    d = (Date) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "date_null_init");
    assertNull(d);

    // date_empty_init
    d = (Date) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "date_empty_init");
    assertNull(d);

    // date_init
    d = (Date) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "date_init");
    assertNotNull(d);
    assertEquals(DateUtil.parseDate("2008/07/31/14/00/00"), d);
  }

  private void checkPerformerVariable(QueryRuntimeAPI queryRuntimeAPI, ProcessInstanceUUID instanceUUID) throws BonitaException {
    // perf_no_init
    String p = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "perf_no_init");
    assertNull(p);

    // perf_null_init
    p = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "perf_null_init");
    assertNull(p);

    // perf_empty_init
    p = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "perf_empty_init");
    assertNull(p);

    // perf_init
    p = (String) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "perf_init");
    assertNotNull(p);
    assertEquals("user", p);
  }

  private void checkEnumerationVariable(QueryRuntimeAPI queryRuntimeAPI,
      ProcessInstanceUUID instanceUUID) throws BonitaException {
    // enum_init
    Enumeration enumStat = (Enumeration) queryRuntimeAPI.getProcessInstanceVariable(instanceUUID, "enum_init");
    assertNotNull(enumStat);
    assertNotNull(enumStat.getSelectedValue());
    assertEquals("iiii", enumStat.getSelectedValue());

  }
}
